#define	NDEBUG

#include <gnome.h>
#include <stdio.h>
#include <string.h>
#include <errno.h>
#include <stdlib.h>
#include <stdarg.h>
#include <sys/types.h>
#include <sys/wait.h>
#include <unistd.h>
#include <fcntl.h>
#include <signal.h>

#include "backend.h"
#include "gicid.h"

static int	debug = 0;
static char	*default_ici_err = "ICI error";

#ifdef __GUNC__
inline
#endif
static int
argv_size(char **argv, int *sz)
{
    int	i;

    for (*sz = i = 0; *argv != NULL; ++i)
	*sz += sizeof (char *) + strlen(*argv++) + 1;
    return i+1;
}

static int
blocking(Backend *b, int block)
{
    int	rc;

    if (!block)
	rc = fcntl(b->fromfd, F_SETFL, O_NONBLOCK);
    else
    {
	int flags;

	if ((rc = fcntl(b->fromfd, F_GETFL, &flags)) != -1)
	    rc = fcntl(b->fromfd, F_SETFL, flags &= ~O_NONBLOCK);
    }
    if (rc == -1)
	fprintf(stderr, "Warning: could not reset non-blocking I/O: %s\n", strerror(errno));
    return rc == -1;
}

Backend *
Backend_new(char **args)
{
    int		p1[2], p2[2];
    char	envvar[32];
    Backend	*b;
    char	**argv;
    char	*p;
    int		i, n, sz;

    if (getenv("GICID_DEBUG"))
	debug = atoi(getenv("GICID_DEBUG"));
    signal(SIGPIPE, SIG_IGN);
    if ((b = malloc(sizeof (Backend))) == NULL)
	return NULL;
    memset(b, 0, sizeof (Backend));
    b->pid = -1;
    b->state = Starting;
    if (pipe(p1) == -1)
    {
	b->state = Broken;
	b->err = errno;
	return b;
    }
    if (pipe(p2) == -1)
    {
	b->err = errno;
	b->state = Broken;
	close(p1[0]);
	close(p1[1]);
	return b;
    }
    switch (b->pid = fork())
    {
    case -1:
	b->state = Broken;
	b->err = errno;
	close(p1[0]);
	close(p1[1]);
	close(p2[0]);
	close(p2[1]);
	return b;

    default:
	close(p1[0]);
	b->tofd = p1[1];
	close(p2[1]);
	b->fromfd = p2[0];
	blocking(b, FALSE);
	if ((b->fromf = fdopen(b->fromfd, "r")) == NULL)
	{
	    Backend_interrupt(b);
	    usleep(100);
	    Backend_interrupt(b);
	    free(b);
	    return NULL;
	}
	setvbuf(b->fromf, NULL, _IONBF, 0);
	b->inputid = 0;
	return b;

    case 0:
	break;
    }

    /* Child */
    close(0);
    open("/dev/null", O_RDONLY);
    close(p1[1]);
    close(p2[0]);
    sprintf(envvar, "ICID_INFD=%d", p1[0]);
    putenv(envvar);
    sprintf(envvar, "ICID_OUTFD=%d", p2[1]);
    putenv(envvar);

    n = argv_size(args, &sz);
    sz += sizeof (char *) + strlen(GICID_CMD) + 1;
    if ((argv = malloc(sz)) == NULL)
    {
	perror("malloc");
	_exit(1);
    }
    p = (char *)&args[n+1];
    argv[0] = strcpy(p, GICID_CMD);
    if (debug)
	fprintf(stderr, "exec: %s", argv[0]);
    p += strlen(p) + 1;
    for (i = 0; i < n-1; ++i)
    {
	argv[i+1] = strcpy(p, args[i]);
	if (debug)
	    fprintf(stderr, " %s", p);
	p += strlen(p) + 1;
    }
    if (debug)
	fprintf(stderr, "\n");
    argv[n] = NULL;
    execvp(argv[0], argv);
    perror(argv[0]);
    _exit(1);
}

int
Backend_sync(Backend *b)
{
    int		nb;
    char	*p;

    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
    {
	nb = -1;
	if (!ferror(b->fromf))
	    ++nb;
    }
    else
    {
	nb = strlen(b->buf);
    }
    if (nb < 0)
    {
	b->state = Broken;
	b->err = errno;
	return 1;
    }
    if (nb == 0)
    {
	b->state = Broken;
	b->err = EIO;
	return 1;
    }
    if (debug)
	fprintf(stderr, "from: '%.*s'\n", nb-1, b->buf);
    switch (b->buf[0])
    {
    case PROGRAM_ERROR:
	b->state = Errored;
	if ((b->icierr = strdup(b->buf + 2)) == NULL)
	    b->icierr = default_ici_err;
	break;

    case INTERNAL_ERROR:
	b->state = Broken;
	b->err = EINVAL;
	break;

    case USER_ERROR:
	break;

    case RETURNED_VALUE:
	b->state = Running;
	break;

    case STOPPED:
	b->file = b->buf + 2;
	p = strrchr(b->file, ' ');
	*p++ = 0;
	b->line = atoi(p);
	if
	(
	    b->state == Starting
	    ||
	    b->state == Running
	    ||
	    b->state == Breakpoint
	    ||
	    b->state == Watchpoint
	    ||
	    b->state == Errored
	)
	{
	    b->state = Stopped;
	}
	break;

    case BROKEN_PROGRAM:
	b->state = Broken;
	if ((b->icierr = strdup(b->buf + 2)) == NULL)
	    b->icierr = default_ici_err;
	break;

    case PROGRAM_EXIT:
	b->status = atoi(b->buf + 2);
	b->state = Finished;
	break;

    case PROGRAM_FINISHED:
	break;

    case BREAKPOINT_HIT:
	b->state = Breakpoint;
	break;

    case WATCHPOINT_HIT:
	b->state = Watchpoint;
	break;
    }
    return 0;
}

static int
send_cmd(Backend *b, int cmd, char *fmt, ...)
{
    char	buf[1024];
    va_list	va;
    int		len;
    int		nb;

    buf[0] = cmd;
    buf[1] = ' ';
    va_start(va, fmt);
    len = 2 + vsprintf(buf + 2, fmt, va);
    va_end(va);
    buf[len++] = '\n';
    buf[len]   = '\0';
    if (debug)
	fprintf(stderr, "  to: '%.*s'\n", len-1, buf);
    if ((nb = write(b->tofd, buf, len)) != len)
    {
	b->state = Broken;
	if (nb == -1)
	    b->err = errno;
	else
	    b->err = EIO;
	return 1;
    }
    return 0;
}

int
Backend_step(Backend *b)
{
    if (send_cmd(b, SINGLE_STEP_CMD, ""))
	return 1;
    b->state = Running;
    if (b->cb != NULL)
	(*b->cb)(b);
    return 0;
}

int
Backend_next(Backend *b)
{
    if (send_cmd(b, STEP_OVER_CMD, ""))
	return 1;
    b->state = Running;
    if (b->cb != NULL)
	(*b->cb)(b);
    return 0;
}

int
Backend_finish(Backend *b)
{
    if (send_cmd(b, FINISH_FUNC_CMD, ""))
	return 1;
    b->state = Running;
    if (b->cb != NULL)
	(*b->cb)(b);
    return 0;
}

int
Backend_continue(Backend *b)
{
    if (send_cmd(b, CONTINUE_CMD, ""))
	return 1;
    b->state = Running;
    if (b->cb != NULL)
	(*b->cb)(b);
    return 0;
}

int
Backend_interrupt(Backend *b)
{
    return kill(b->pid, SIGINT) == -1;
}

void
Backend_destroy(Backend *b)
{
    if (b->inputid != 0)
	gtk_input_remove(b->inputid);
    kill(b->pid, SIGTERM);
    close(b->tofd);
    close(b->fromfd);
    if (b->icierr != NULL && b->icierr != default_ici_err)
	free(b->icierr);
    free(b);
}

static void
backend_input_func(gpointer data, gint source, GdkInputCondition cond)
{
    (void)source;
    (void)cond;
    Backend_sync((Backend *)data);
    if (((Backend *)data)->cb != NULL)
	(*((Backend *)data)->cb)((Backend *)data);
}

int
Backend_listen(Backend *b, void (*cb)(Backend *))
{
    b->inputid = gtk_input_add_full
    (
	b->fromfd, GDK_INPUT_READ, backend_input_func,
	NULL, b, NULL
    );
    b->cb = cb;
    return 0;
}

int
Backend_where(Backend *b, char ***lines, int *nlevels)
{
    int		i;

    if (send_cmd(b, WHERE_CMD, ""))
	return 1;
    blocking(b, TRUE);
    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	goto fail;
    if (*b->buf != STACK_FRAME)
	goto fail;
    if (sscanf(b->buf + 2, "%d", nlevels) != 1)
	goto fail;
    if ((*lines = malloc(*nlevels * sizeof (char *))) == NULL)
	goto fail;
    for (i = 0; i < *nlevels; ++i)
    {
	if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	    goto fail;
	if (((*lines)[i] = malloc(strlen(b->buf) + 1)) == NULL)
	    goto fail2;
	strcpy((*lines)[i], b->buf);
    }
    blocking(b, FALSE);
    return 0;

fail2:
    free(*lines);
fail:
    blocking(b, FALSE);
    return 1;
}

void
Backend_free_where(char **lines, int nlines)
{
    int		i;

    for (i = 0; i < nlines; ++i)
	free(lines[i]);
    free(lines);
}

void *
Backend_get_scope(Backend *b)
{
    void	*p;

    if (send_cmd(b, GET_SCOPE_CMD, ""))
	return NULL;
    if (blocking(b, TRUE))
	return NULL;
    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	return NULL;
    if (blocking(b, FALSE))
	return NULL;
    if (*b->buf != SCOPE)
	return NULL;
    if (sscanf(b->buf+2, "%p", &p) != 1)
	return NULL;
    return p;
}

int
Backend_set_scope(Backend *b, void *scp)
{
    if (send_cmd(b, SET_SCOPE_CMD, "%p", scp))
	return 1;
    if (blocking(b, TRUE))
	return 1;
    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	return 1;
    if (blocking(b, FALSE))
	return 1;
    return *b->buf != OKAY;
}

int
Backend_scope_up(Backend *b)
{
    if (send_cmd(b, SCOPE_UP_CMD, ""))
	return 1;
    if (blocking(b, TRUE))
	return 1;
    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	return 1;
    if (blocking(b, FALSE))
	return 1;
    return *b->buf != OKAY;
}

int
Backend_fetch(Backend *b, const char *key, char *val, int valsz)
{
    char	*p;

    if (send_cmd(b, FETCH_CMD, "%s", key))
	return 1;
    if (blocking(b, TRUE))
	return 1;
    for (p = val;;)
    {
	int	len;

	if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	    return 1;
	if (b->buf[0] == FETCHED_VALUE && b->buf[1] == ' ' && b->buf[2] == '\n')
	    break;
	len = strlen(b->buf + 2);
	if ((p + len + 1) < (val + valsz))
	{
	    char	*np;

	    strcpy(p, b->buf + 2);
	    if ((np = strrchr(p, '\n')) == NULL)
		p += len;
	    else
	    {
		*np = ' ';
		p = np + 1;
	    }
	    *p = '\0';
	}
    }
    if (blocking(b, FALSE))
	return 1;
    return 0;
}

int
Backend_assign(Backend *b, const char *key, const char *val)
{
    if (send_cmd(b, ASSIGN_CMD, "%s %s", key, val))
	return 1;
    if (blocking(b, TRUE))
	return 1;
    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	return 1;
    if (blocking(b, FALSE))
	return 1;
    return *b->buf == OKAY;
}

int
Backend_scope_keys(Backend *b, char ***keys, int *nkeys)
{
    int	i;

    if (send_cmd(b, SCOPE_KEYS_CMD, ""))
	return 1;
    blocking(b, TRUE);
    if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	goto fail;
    if (*b->buf != SCOPE_KEYS)
	goto fail;
    if (sscanf(b->buf + 2, "%d", nkeys) != 1)
	goto fail;
    if ((*keys = malloc(*nkeys * sizeof (char *))) == NULL)
	goto fail;
    for (i = 0; i < *nkeys; ++i)
    {
	char	*p;
	if (fgets(b->buf, sizeof b->buf, b->fromf) == NULL)
	    goto fail;
	if ((p = strrchr(b->buf, '"')) != NULL)
	    *p = '\0';
	if (((*keys)[i] = malloc(strlen(b->buf+1) + 1)) == NULL)
	    goto fail2;
	strcpy((*keys)[i], b->buf+1);
    }
    blocking(b, FALSE);
    return 0;

fail2:
    free(*keys);
fail:
    blocking(b, FALSE);
    return 1;
}
