#include <gnome.h>
#include <stdio.h>
#include <stdlib.h>
#include <stdarg.h>
#include <string.h>
#include <errno.h>
#include <sys/types.h>
#include <signal.h>
#include <unistd.h>

#include "interface.h"
#include "fe.h"
#include "backend.h"
#include "main.h"
#include "dialog.h"
#include "filecache.h"

static char	**remembered_args = NULL;
static int	dialog_state	  = -1;
static char	currentfname[FILENAME_MAX];
static int	file;
static void	*curscp = 0;
static void	**scpstk;
static int	nscp;
static int	maxscp;

static int
okay_clicked(gpointer p)
{
    gtk_widget_hide(dialog);
    gtk_main_quit();
    gtk_widget_destroy(dialog);
    dialog_state = 1;
    return 0;
}

static int
cancel_clicked(gpointer p)
{
    gtk_widget_hide(dialog);
    gtk_main_quit();
    gtk_widget_destroy(dialog);
    dialog_state = 0;
    return 0;
}

int
do_dialog(int nbuttons, const char *title, const char *msg, ...)
{
    char	buffer[4096];
    va_list	va;
    GtkWidget	*w;

    va_start(va, msg);
    vsprintf(buffer, msg, va);
    va_end(va);

    dialog_state = -1;
    dialog = make_dialog(nbuttons);
    w = fwidget(dialog, "frame");
    gtk_frame_set_label(GTK_FRAME(w), title);
    w = fwidget(dialog, "label");
    gtk_label_set_text(GTK_LABEL(w), buffer);
    w = fwidget(dialog, "button1");
    gtk_signal_connect(GTK_OBJECT(w), "clicked", GTK_SIGNAL_FUNC(okay_clicked), dialog);
    if ((w = fwidget(dialog, "button2")) != NULL)
	gtk_signal_connect(GTK_OBJECT(w), "clicked", GTK_SIGNAL_FUNC(cancel_clicked), dialog);
    gtk_widget_show_all(dialog);
    gtk_main();
    return dialog_state;
}

int
get_dialog_state(void)
{
    return dialog_state;
}

static void
file_open1(const char *fname)
{
    GtkText		*text;
    static GdkFont	*font = NULL;
    int			len;
    int			i;
    char		*line;

    if ((file = FileCache_open(fname)) == -1)
    {
	do_dialog
	(
	    1,
	    "File Open Failed",
	    "Unable to open,\n"
	    "\n"
	    "\t%s\n"
	    "\n"
	    "%s",
	    fname,
	    strerror(errno)
	);
	return;
    }
    if (font == NULL)
    {
	char	*fontname;

	if ((fontname = getenv("FONT")) == NULL)
	    fontname = "-*-lucidatypewriter-medium-r-normal-*-*-110-*-*-*-*-*-*";
	font = gdk_font_load(fontname);
    }
    text = GTK_TEXT(fwidget(window, "text"));
    gtk_text_freeze(text);
    gtk_text_set_editable(text, TRUE);
    gtk_text_set_point(text, 0);
    len = gtk_text_get_length(text);
    gtk_text_forward_delete(text, len);
    len = FileCache_nlines(file);
    for (i = 0; i < len; )
    {
	line = FileCache_line(file, ++i);
	gtk_text_insert(text, font, NULL, NULL, line, strlen(line));
	gtk_text_set_point(text, gtk_text_get_length(text));
    }
    strcpy(currentfname, fname);
    gtk_text_set_editable(text, FALSE);
    gtk_text_thaw(text);
    gtk_frame_set_label(GTK_FRAME(fwidget(window, "frame10")), fname);
}

void
file_open(const char *fname)
{
    file_open1(fname);
    restart_program();
}

void
remember_args(char **args)
{
    remembered_args = args;
}

char **
get_remembered_args(void)
{
    return remembered_args;
}

static void
highlight_line(const char *fname, int lineno)
{
    GtkText		*text;
    int			pos, end;
    char		*line;
    GtkAdjustment	*adj;
    double		v;

    if (*fname == '\0' || strcmp(fname, ".") == 0)
	return;
    if (strcmp(fname, currentfname) != 0)
	file_open1(fname);
    if (file == -1)
	return;
    text = GTK_TEXT(fwidget(window, "text"));
    pos = FileCache_linepos(file, lineno);
    line = FileCache_line(file, lineno);
    end = pos + strlen(line);
    while (*line && (*line == ' ' || *line == '\t'))
    {
	++pos;
	++line;
    }
    adj = gtk_scrolled_window_get_vadjustment
    (
	GTK_SCROLLED_WINDOW(fwidget(window, "scrolledwindow8"))
    );
    v = ((double)lineno / FileCache_nlines(file)) * (adj->upper - adj->lower + 1);
    v += 3.0 * adj->page_size / 4.0;
    gtk_adjustment_clamp_page(adj, v - adj->page_size, v);
    gtk_adjustment_changed(adj);
    gtk_editable_select_region(GTK_EDITABLE(text), pos, end);
}

static void
update_backend_status(Backend *b)
{
    char	msg[1024];

    switch (b->state)
    {
    case Running:
	gnome_appbar_set_status(GNOME_APPBAR(fwidget(window, "appbar1")), "Running");
	enable_buttons(FALSE, FALSE, TRUE);
	break;

    case Stopped:
	highlight_line(b->file, b->line);
	sprintf(msg, "Stopped at line %d", b->line);
	gnome_appbar_set_status(GNOME_APPBAR(fwidget(window, "appbar1")), msg);
	enable_buttons(TRUE, TRUE, FALSE);
	update_scope_view();
	update_stack_frames();
	break;

    case Finished:
	sprintf(msg, "Program finished, exit status = %d", b->status);
	gnome_appbar_set_status(GNOME_APPBAR(fwidget(window, "appbar1")), msg);
	enable_buttons(TRUE, FALSE, FALSE);
	break;

    case Errored:
	do_dialog
	(
	    1,
	    "Error",
	    b->icierr
	);
	enable_buttons(TRUE, TRUE, FALSE);
	break;

    case Breakpoint:
	sprintf(msg, "Breakpoint hit");
	gnome_appbar_set_status(GNOME_APPBAR(fwidget(window, "appbar1")), msg);
	enable_buttons(TRUE, TRUE, FALSE);
	break;

    case Watchpoint:
	sprintf(msg, "Watchpoint hit");
	gnome_appbar_set_status(GNOME_APPBAR(fwidget(window, "appbar1")), msg);
	enable_buttons(TRUE, TRUE, FALSE);
	break;

    case Broken:
	enable_buttons(TRUE, FALSE, FALSE);
	ask_restart_program();
	break;
    }
}

void
restart_program(void)
{
    if ((backend = Backend_new(get_remembered_args())) == NULL)
    {
	do_dialog
	(
	    1,
	    "Failed To Restart Program!",
	    "Unable to restart program."
	);
	gtk_main_quit();
    }
    else
    {
	Backend_listen(backend, update_backend_status);
    }
}

void
ask_restart_program(void)
{
    if
    (
	do_dialog
	(
	    2,
	    "Restart Program?",
	    "The program has finished. Press \"Okay\" if you "
	    "would like to restart the program otherwise press "
	    "\"Cancel\""
	)
    )
    {
	Backend_destroy(backend);
	restart_program();
    }
}

void
enable_buttons(int runb, int stepb, int intrb)
{
    GtkWidget	*toolbar = fwidget(window, "toolbar3");

    gtk_widget_set_sensitive(fwidget(toolbar, "button11"), runb);
    gtk_widget_set_sensitive(fwidget(toolbar, "button8"), stepb);
    gtk_widget_set_sensitive(fwidget(toolbar, "button9"), stepb);
    gtk_widget_set_sensitive(fwidget(toolbar, "button10"), stepb);
    gtk_widget_set_sensitive(fwidget(toolbar, "button12"), intrb);
}

static int
cmpkey(const void *a, const void *b)
{
    return strcmp(*(const char **)a, *(const char **)b);
}

void
update_scope_view(void)
{
    GtkCList	*clist;
    char	**keys;
    int		nkeys;
    int		i;
    char	*cols[2];

    if (Backend_set_scope(backend, curscp))
	return;
    if (Backend_scope_keys(backend, &keys, &nkeys))
	return;
    if (nkeys > 0)
	qsort(keys, nkeys, sizeof (char *), cmpkey);
    if ((clist = GTK_CLIST(fwidget(window, "clist"))) == NULL)
	goto ret;
    gtk_clist_freeze(clist);
    gtk_clist_clear(clist);
    for (i = 0; i < nkeys; ++i)
    {
	char	val[10240];

	if (strcmp(keys[i], "_funcname_") == 0)
	    continue;
	cols[0] = keys[i];
	cols[1] = val;
	if (Backend_fetch(backend, keys[i], val, sizeof val) == 0)
	    gtk_clist_append(clist, cols);
    }
    gtk_clist_thaw(clist);
ret:
    Backend_free_where(keys, nkeys);
}

static int
pushscp(void *scp)
{
    if (nscp == maxscp)
    {
	void	**newstk;
	if (maxscp == 0)
	    maxscp = 16;
	else if (maxscp < 64 * 1024)
	    maxscp *= 2;
	else
	    maxscp += 1024;
	if ((newstk = realloc(scpstk, maxscp * sizeof scpstk[0])) == NULL)
	    return 1;
	scpstk = newstk;
    }
    scpstk[nscp++] = scp;
    return 0;
}

static void *
popscp(void)
{
    void	*rc;

    if (nscp == 0)
	return NULL;
    rc = scpstk[--nscp];
    if (nscp < maxscp/2)
    {
	void	**newstk;

	if ((newstk = realloc(scpstk, maxscp/2 * sizeof scpstk[0])) != NULL)
	{
	    scpstk = newstk;
	    maxscp /= 2;
	}
    }
    return rc;
}

static void
update_scope_label(void)
{
    char	*txt[] = {"Autos", "Statics", "Externs"};
    GtkFrame	*w = GTK_FRAME(fwidget(window, "frame8"));

    if (w != NULL)
    {
	if (nscp <= 2)
	    gtk_frame_set_label(w, txt[nscp]);
	else
	{
	    char	buf[32];

	    sprintf(buf, "Externs + %d", nscp - 2);
	    gtk_frame_set_label(w, buf);
	}
    }
}

void
scope_view_up(void)
{
    if (Backend_scope_up(backend) || pushscp(curscp))
	return;
    curscp = Backend_get_scope(backend);
    update_scope_view();
    update_scope_label();
}

void
scope_view_down(void)
{
    curscp = popscp();
    update_scope_view();
    update_scope_label();
}

void
update_stack_frames(void)
{
    char	**lines;
    int		nlines;
    int		i;
    GtkCList	*clist;

    if (Backend_where(backend, &lines, &nlines))
	return;
    if ((clist = GTK_CLIST(fwidget(window, "calllist"))) == NULL)
	goto ret;
    gtk_clist_freeze(clist);
    gtk_clist_clear(clist);
    for (i = nlines - 1; i >= 0; --i)
    {
	char	*clines[2];
	char	level[16];

	sprintf(level, "%d", i+1);
	clines[0] = level;
	clines[1] = lines[i];
	gtk_clist_append(clist, clines);
    }
    gtk_clist_thaw(clist);
ret:
    Backend_free_where(lines, nlines);
}
