/*
 * $Id: gicid.c,v 1.6 2001/03/04 05:23:50 andy Exp $
 *
 * Copyright (C) 2001.   A.Newman.
 * All Rights Reserved.
 *
 */

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>

#include <object.h>
#include <int.h>
#include <float.h>
#include <str.h>
#include <re.h>
#include <array.h>
#include <set.h>
#include <struct.h>
#include <ptr.h>
#include <src.h>
#include <buf.h>
#include <func.h>
#include <exec.h>
#include <op.h>
#include <file.h>
#include <null.h>

#include "gicid.h"

/*
 * The different states we can be in. This is used to decide when we
 * should stop execution and prompt the user for what to do next. The
 * ici interpreter calls the debug callback functions at specific points
 * and these functions need to decide what to do according to the state.
 * (The reason for actually naming the enum and using this is so debuggers
 *  will be nice to us and print the enumerators, sad but true).
 */
enum state
{
    FINISHED,		/* Program has finished execution.		*/
    SINGLE_STEP,	/* Single-step, stop at each statement.		*/
    SINGLE_STEP_OVER,	/* Single-step, don't step through functions.	*/
    RUN_TILL_BREAK,	/* Run until a break- or watch-point is hit.	*/
    RUN_TILL_RETURN,	/* Run until the current function returns.	*/
    BROKEN		/* The program raised an unhandled error.	*/
};

FILE		*finput;
FILE		*foutput;
int		dontstop;
enum state	state = SINGLE_STEP;
struct_t	*watched;
object_t	**tos;
src_t		*srcmrk;
struct_t	*breakpoints;
struct_t	*breakfuncs;
array_t		*callstack;
volatile int	interrupt;
string_t	*startupfile;
int		needinit = 1;
struct_t	*sources;
string_t	*string__;
string_t	*string_id;
string_t	*string_where;
string_t	*string_Type;
string_t	*string_func;
string_t	*string_line;
char		*floatfmt = "%*s%.8g";
int		maxstrlen = 32;
int		maxarraylen = 3;
int		breaknum = 0;
struct_t	*curscp = NULL;

#define	min(a,b) ((a) < (b) ? (a) : (b))
#define	max(a,b) ((a) > (b) ? (a) : (b))

typedef struct
{
    char	cmdname;
    int		resume;
    void	(*func)(char *);
}
cmd_t;

static void	breakpt(char *);
static void	cont(char *);
static void	fin(char *);
static void	next(char *);
static void	print(char *);
static void	quit(char *);
static void	step(char *);
static void	watch(char *);
static void	where(char *);
static void	getscope(char *);
static void	setscope(char *);
static void	scopeup(char *);
static void	fetchcmd(char *);
static void	assigncmd(char *);
static void	scopekeys(char *);

static cmd_t	cmds[] =
{
    {SET_BREAKPOINT_CMD,	0,	breakpt},
    {CONTINUE_CMD,		1,	cont},
    {FINISH_FUNC_CMD,		1,	fin},
    {STEP_OVER_CMD,		1,	next},
    {PRINT_VALUE_CMD,		0,	print},
    {QUIT_CMD,			0,	quit},
    {SINGLE_STEP_CMD,		1,	step},
    {SET_WATCHPOINT_CMD,	0,	watch},
    {WHERE_CMD,			0,	where},
    {GET_SCOPE_CMD,		0,	getscope},
    {SET_SCOPE_CMD,		0,	setscope},
    {SCOPE_UP_CMD,		0,	scopeup},
    {FETCH_CMD,			0,	fetchcmd},
    {ASSIGN_CMD,		0,	assigncmd},
    {SCOPE_KEYS_CMD,		0,	scopekeys},
};

#define		ncmds (sizeof cmds/sizeof cmds[0])

static void	reply(const int, const char *, ...);
#define	init()	if (needinit && init_()) fatal(error)
static int	init_(void);
static void	stopped(void);
static void	sigint(int);
#define	fatal(s) fatal_(s, __FILE__, __LINE__)
static void	fatal_(char *, char *, int);
static cmd_t	*search(char *);
static void	pobj(object_t *, int, int);
static void	enterstate(enum state);

void		ici_debug_error(char *, src_t *);
void		ici_debug_fncall(object_t *, object_t **, int);
void		ici_debug_fnresult(object_t *);
void		ici_debug_src(src_t *);
void		ici_debug_watch(object_t *, object_t *, object_t *);

debug_t		icid_debug_funcs =
{
    ici_debug_error,
    ici_debug_fncall,
    ici_debug_fnresult,
    ici_debug_src,
    ici_debug_watch
};

int
main(int argc, char **argv)
{
    int		i;

    ici_debug = &icid_debug_funcs;
    finput = stdin;
    foutput = stdout;
    if (getenv("ICID_INFD") != NULL)
    {
	finput = fdopen(atoi(getenv("ICID_INFD")), "r");
	foutput = fdopen(atoi(getenv("ICID_OUTFD")), "w");
	setlinebuf(foutput);
	setlinebuf(finput);
    }
    signal(SIGINT, sigint);
    i = ici_maind(argc, argv, 1);
    dontstop = 1;
    reply(PROGRAM_EXIT, "%d", i);
    enterstate(FINISHED);
    stopped();
    return 0;
}

/*
 * Send a reply message to our master.
 */
static void
reply(const int what, const char *msg, ...)
{
    va_list	va;

    fprintf(foutput, "%c", what);
    if (msg != NULL)
    {
	fprintf(foutput, " ");
	va_start(va, msg);
	vfprintf(foutput, msg, va);
	va_end(va);
    }
    fprintf(foutput, "\n");
    fflush(foutput);
}

/*
 * Version of exit() using _exit(). Flush all files before exiting.
 */
static void
my_exit(int rc)
{
    fflush(NULL);
    _exit(rc);
}

#ifdef __GNUC__
static void	my_exit(int) __attribute__ ((noreturn));
#endif

/*
 * This is gnarly. Override exit so we can trap ici calls to it.
 * The code is a copy of that in main() following the call to
 * ici_maind().
 */
void
exit(int rc)
{
    if (!dontstop)
    {
	reply(PROGRAM_EXIT, "%d", rc);
	enterstate(FINISHED);
	stopped();
    }
    my_exit(rc);
}

/*
 * Replace exit() with my_exit() in the remainder of this file so
 * we don't call the above function when we try to exit the
 * debugger itself.
 */
#define	exit(s)	my_exit(s)

/*
 * Trap SIGINT (and possibly others) and set the interrupt flag so
 * the callbacks know to stop the program. If more than one SIGINT
 * in a row exit immediately.
 */
static void
sigint(int sig)
{
    if (++interrupt > 1)
    {
	reply(INTERNAL_ERROR, "double sigint");
	exit(1);
    }
}

/*
 * We define a new type, callrec, to record details of function calls.
 * We maintain an array of callrec objects to record the function call
 * "stack frames". This allows us to implement the "where" command.
 *
 * Each callrec contains the parameters to ici_debug_fncall (see below)
 * at the time of the function call. The argument pointer array, ap, is
 * a copy of that passed to ici_debug_fncall as the parameter points
 * directly into the ici stack which may change and has a "backwards"
 * layout in memory.
 *
 * XXX * THE VALUES THE OBJECTS REPRESENT MAY CHANGE AND THE DEBUGGER
 * SHOULD REALLY DO A DEEP COPY OF THE ARGUMENTS SO IT CAN REPORT THE
 * INITIAL VALUES CORRECTLY. HOWEVER I'LL LEAVE THAT OUT FOR NOW.
 */
typedef struct callrec
{
    object_t	o_head;
    object_t	*fn;
    object_t	**ap;
    int		nargs;
}
callrec_t;

#define	callrecof(o)	((callrec_t *)(o))

static unsigned long
mark_callrec(object_t *o)
{
    unsigned long	sz;
    int			i;

    o->o_flags |= O_MARK;
    for (sz = 0, i = 0; i < callrecof(o)->nargs; ++i)
	sz += mark(callrecof(o)->ap[i]);
    return sizeof (callrec_t) + mark(callrecof(o)->fn) + sz;
}

static type_t	callrec_type =
{
    mark_callrec,
    free_simple,
    hash_unique,
    cmp_unique,
    copy_simple,
    assign_simple,
    fetch_simple,
    "callrec"
};

static callrec_t *
new_callrec(object_t *fn, object_t **ap, int nargs)
{
    size_t	sz;
    callrec_t	*cr;

    sz = nargs * sizeof (object_t *);
    if ((cr = ici_alloc(sizeof (callrec_t) + sz)) != NULL)
    {
	int	i;

	objof(cr)->o_tcode = TC_OTHER;
	objof(cr)->o_flags = 0;
	objof(cr)->o_nrefs = 1;
	objof(cr)->o_type  = &callrec_type;
	rego(cr);
	cr->fn = fn;
	cr->ap = (object_t **)(cr + 1);
	cr->nargs = nargs;
	for (i = 0; i < nargs; ++i)
	{
	    cr->ap[i] = ap[-i]; /* Note: ap is backwards, see ici/func.h */
	    incref(cr->ap[i]);
	}
    }
    return cr;
}

/****
 * Utility functions...
 */

static void
fatal_(char *msg, char *file, int line)
{
    reply(INTERNAL_ERROR, "%s:%d: %s", file, line, msg);
    exit(1);
}

static int
broken(void)
{
    reply(BROKEN_PROGRAM, NULL);
    return 1;
}

static void
finished(void)
{
    reply(PROGRAM_FINISHED, NULL);
}

/*
 * This is set in the object flags if this object has been output.
 * We don't want to get into infinite loops when we output things.
 * That would be bad.
 */
#define	O_PRINT	0x20

/*
 * Output a string or regular expression object.
 */
static void
pstrorreg(object_t *o, int ind)
{
    int	i, n, c;

    n = min(maxstrlen, stringof(o)->s_nchars);
    fprintf(foutput, "%*s%c", ind, "", isstring(o) ? '"' : '#');
    for (i = 0; i < n; ++i)
    {
	switch (c = stringof(o)->s_chars[i])
	{
	case '\n':
	    fprintf(foutput, "\\n");
	    break;
	case '\r':
	    fprintf(foutput, "\\r");
	    break;
	case '\b':
	    fprintf(foutput, "\\b");
	    break;
	case '\v':
	    fprintf(foutput, "\\v");
	    break;
	case '\f':
	    fprintf(foutput, "\\f");
	    break;
	default:
	    if (c >= ' ' && c <= '~')
		fputc(c, foutput);
	    else
		fprintf(foutput, "\\x%02X", c);
	}
    }
    fputc(isstring(o) ? '"' : '#', foutput);
    if (stringof(o)->s_nchars > maxstrlen)
	fprintf(foutput, "...");
}

/*
 * Output a pointer
 */
static void
pptr(object_t *o, int ind1, int ind2)
{
    fprintf(foutput, "%*s&", ind1, "");
    pobj(ptrof(o)->p_aggr, 0, ind2);
    fprintf(foutput, "[");
    pobj(ptrof(o)->p_key, 0, ind2);
    fprintf(foutput, "]");
}

/*
 * Output an array
 */
static void
parray(object_t *o, int ind1, int ind2)
{
    object_t	**po;
    int		i;

    o->o_flags |= O_PRINT;
    fprintf(foutput, "%*s%s[array\n", ind1, "", o->o_flags & O_ATOM ? "@" : "");
    for
    (
	i = 0, po = arrayof(o)->a_base;
	i < maxarraylen && po < arrayof(o)->a_top;
	++po, ++i
    )
    {
	fprintf(foutput, "%*s[%d] = ", ind2 + 1, "", i);
	pobj(*po, 0, ind2 + 1);
	fprintf(foutput, ",\n");
    }
    if (po < arrayof(o)->a_top)
	fprintf(foutput, "%*s...\n", ind2 + 1, "");
    fprintf(foutput, "%*s]", ind2, "");
    o->o_flags &= ~O_PRINT;
}

/*
 * Output a structure
 */
static void
pstruct(object_t *o, int ind1, int ind2)
{
    slot_t	*sl;

    o->o_flags |= O_PRINT;
    fprintf(foutput, "%*s%s[struct\n", ind1, "", o->o_flags & O_ATOM ? "@" : "");
    for
    (
	sl = structof(o)->s_slots;
	(sl - structof(o)->s_slots) < structof(o)->s_nslots;
	++sl
    )
    {
	if (sl->sl_key != NULL)
	{
	    pobj(sl->sl_key, ind2 + 1, ind2 + 1);
	    fprintf(foutput, " = ");
	    pobj(sl->sl_value, 0, ind2 + 1);
	    fprintf(foutput, ",\n");
	}
    }
    fprintf(foutput, "%*s]", ind2, "");
    o->o_flags &= ~O_PRINT;
}

/*
 * Output a set.
 */
static void
pset(object_t *o, int ind1, int ind2)
{
    object_t	**po;

    o->o_flags |= O_PRINT;
    fprintf(foutput, "%*s%s[set\n", ind1, "", o->o_flags & O_ATOM ? "@" : "");
    for
    (
	po = setof(o)->s_slots;
	po - setof(o)->s_slots < setof(o)->s_nslots;
	++po
    )
    {
	if (*po != NULL)
	{
	    pobj(*po, ind2 + 1, ind2 + 1);
	    fprintf(foutput, ",\n");
	}
    }
    fprintf(foutput, "%*s]", ind2, "");
    o->o_flags &= ~O_PRINT;
}

#ifdef ici3
#define	func_name(f) ((f)->f_name)
#endif

/*
 * Print an ici object to the standard output.
 */
static void
pobj(object_t *o, int ind1, int ind2)
{
    if (o->o_flags & O_PRINT)
	return;
    if (isnull(o))
	fprintf(foutput, "%*sNULL", ind1, "");
    else if (isint(o))
	fprintf(foutput, "%*s%ld", ind1, "", intof(o)->i_value);
    else if (isfloat(o))
	fprintf(foutput, floatfmt, ind1, "", floatof(o)->f_value);
    else if (isstring(o) || isregexp(o))
	pstrorreg(o, ind1);
    else if (isptr(o))
	pptr(o, ind1, ind2);
    else if (isarray(o))
	parray(o, ind1, ind2);
    else if (isset(o))
	pset(o, ind1, ind2);
    else if (isstruct(o))
	pstruct(o, ind1, ind2);
    else if (issrc(o))
	fprintf(foutput, "%s:%d", srcof(o)->s_filename->s_chars, srcof(o)->s_lineno);
    else if (isfunc(o))
	fprintf(foutput, "%*s%s()", ind1, "", func_name(funcof(o))->s_chars);
    else
	fprintf(foutput, "%*s[%s]", ind1, "", o->o_type->t_name);
}

/*
 * Execute ici code stored formed from a printf format string and args.
 */
static object_t *
iexpr(char *fmt, ...)
{
    va_list		vargs;
    char		fmtstr[1024];
    int			len;
    file_t		*file;
    int			rc;
    object_t		*o;

    if (chkbuf(1024 + strlen(fmt)))
	fatal(error);
    va_start(vargs, fmt);
    sprintf(fmtstr, "auto _=%s;\n", fmt);
    len = vsprintf(buf, fmtstr, vargs);
    va_end(vargs);
    if ((file = ici_sopen(buf, len)) == NULL)
	fatal(error);
    if ((file->f_name = get_cname(buf)) == NULL)
	fatal(error);
    dontstop = 1;
    rc = parse_module(file, structof(v_top[-1]));
    decref(file);
    dontstop = 0;
    if (rc < 0)
	return NULL;
    o = fetch(v_top[-1], string__);
    if (o == objof(&o_null) || o == NULL)
	return objof(&o_null);
    unassign_struct(structof(v_top[-1]), objof(string__));
    NEXT_VSVER;
    return o;
}

/*
 * Set the debugger state with special action on transitions.
 */
static void
enterstate(enum state newstate)
{
    state = newstate;
}

/*
 * See if we should stop at the file/line described by the src_t.
 * Return non-zero if we should stop the program, zero otherwise.
 */
static int
stopat(src_t *src)
{
    object_t	*o;
    int_t	*i;
    int		rc;

    if ((i = new_int(src->s_lineno)) == NULL)
	fatal(error);
    o = fetch(breakpoints, i);
    decref(i);
    rc = !isnull(o);
    if (rc)
	reply(BREAKPOINT_HIT, "%s %d", src->s_filename->s_chars, src->s_lineno);
    return rc;
}

/*
 * Determine if we should stop at a particular function.
 */
static int
stopin(object_t *fn)
{
    int		rc;

    if ((rc = !isnull(fetch(breakfuncs, fn))))
	reply(BREAKPOINT_HIT, "%s", func_name(funcof(fn))->s_chars);
    return rc;
}

/****
 * Command functions...
 */

/*
 * Exit the debugger. Do whatever you need to do to clean up here.
 * At present there's nothing to do but that may change.
 */
static void
quit(char *arg)
{
    (void)arg;
    exit(0);
}

/*
 * Print the result on an expression.
 */
static void
print(char *expr)
{
    static object_t	*prev = NULL;

    if (*expr != '\0')
	prev = iexpr(expr);
    else if (prev == NULL)
    {
	reply(USER_ERROR, "no prior print command");
	return;
    }
    if (prev == NULL)
    {
	char	*p;

	/* Strip filename and line number off error text */
	if ((p = strchr(error, ':')) == NULL)
	    p = error;
	else
	    p += 2; /* There's a space following the colon */
	reply('!', "%s", p);
    }
    else
    {
	reply('=', "%s", expr);
	pobj(prev, 0, 0);
    }
}

/*
 * Continue program execution.
 */
static void
cont(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
	finished();
    else
	enterstate(RUN_TILL_BREAK);
}

/*
 * Single-step the program.
 */
static void
step(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
    {
	finished();
	return;
    }
    enterstate(SINGLE_STEP);
}

/*
 * Single-step but don't trace through called functions.
 */
static void
next(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
    {
	finished();
	return;
    }
    tos = callstack->a_top;
    enterstate(SINGLE_STEP_OVER);
}

/*
 * Finish the current function.
 */
static void
fin(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
    {
	finished();
	return;
    }
    tos = callstack->a_top;
    enterstate(RUN_TILL_RETURN);
}

static struct_t *
breakpt_line(char *filename, int lineno)
{
    int_t	*line;
    struct_t	*s1;
    struct_t	*s2;
    string_t	*n;

    if ((line = new_int(lineno)) == NULL)
	fatal(error);
    if ((n = new_cname(filename)) == NULL)
	fatal(error);
    if ((s2 = structof(fetch(breakpoints, line))) == structof(&o_null))
    {
	if ((s2 = new_struct()) == NULL)
	    fatal(error);
	if (assign(breakpoints, line, s2))
	    fatal(error);
	decref(s2);
    }
    else if (fetch(s2, n) != objof(&o_null))
    {
	reply(USER_ERROR, " already set %s %d", filename, lineno);
	decref(line);
	decref(n);
	return NULL;
    }
    if ((s1 = new_struct()) == NULL)
	fatal(error);
    if (assign(s2, n, s1))
	fatal(error);
    decref(s1);
    decref(n);
    return s1;
}


/*
 * Set a breakpoint at a particular file/line or on a function.
 */
static void
breakpt(char *where)
{
    char	*p;
    object_t	*o;
    int		lineno;
    struct_t	*s = 0;
    string_t	*n;
    int_t	*i;
    string_t	*type;

    /*
     * Do a reverse search to cope with `:'s in filenames.
     */
    if ((p = strrchr(where, ':')) == NULL)
    {
	if ((n = new_cname(where)) == NULL)
	    fatal(error);
	if ((o = fetch(v_top[-1], n)) == objof(&o_null))
	{
	    reply(USER_ERROR, "\"%s\" undefined", where);
	    return;
	}
	if (!isfunc(o))
	{
	    reply(USER_ERROR, "%s is not a function", where);
	    return;
	}
	if ((s = new_struct()) == NULL)
	    fatal(error);
	if (assign(breakfuncs, o, s))
	    fatal(error);
	type = string_func;
    }
    else
    {
	if (sscanf(p+1, "%u", &lineno) != 1 || lineno < 1)
	{
	    reply(USER_ERROR, "invalid breakpoint specification");
	    return;
	}
	if ((s = breakpt_line(where, lineno)) == NULL)
	    return;
	type = string_line;
    }
    decref(s);
    if ((i = new_int(++breaknum)) == NULL || assign(s, string_id, i))
	fatal(error);
    decref(i);
    if ((n = new_cname(where)) == NULL || assign(s, string_where, n))
	fatal(error);
    decref(n);
    if (assign(s, string_Type, type))
	fatal(error);
}

/*
 * Set a watchpoint on a variable.
 */
static void
watch(char *arg)
{
    string_t	*s;
    set_t	*set;
    struct_t	*ss;

    if ((s = new_cname(arg)) == NULL)
	fatal(error);
    ss = structof(v_top[-1]);
    if (find_slot(&ss, objof(s))->sl_key != objof(s))
    {
	reply(USER_ERROR, "no such variable");
	decref(s);
	return;
    }
    if (objof(set = setof(fetch(watched, v_top[-1]))) == objof(&o_null))
    {
	if ((set = new_set()) == NULL || assign(watched, v_top[-1], set))
	    fatal(error);
	decref(set);
    }
    if (assign(set, s, o_one))
	fatal(error);
    decref(s);
}

/*
 * Output the function call stack.
 */
static void
where(char *arg)
{
    int		i, j, n;

    (void)arg;
    if (state == FINISHED)
    {
	finished();
	return;
    }
    if (callstack == NULL || callstack->a_top == callstack->a_base)
    {
	reply(USER_ERROR, "where command issued when not in a called function");
	return;
    }
    n = callstack->a_top - callstack->a_base;
    reply(STACK_FRAME, "%d", n);
    for (i = 0; i < n; ++i)
    {
	char		nam[ICI_OBJNAMEZ];
	object_t	*fn;

	fn = callrecof(callstack->a_base[i])->fn;
	if (iscfunc(fn))
	{
	    int	len;
	    (*fn->o_type->t_objname)(fn, nam);
	    len = strlen(nam);
	    if (len > 2)
		nam[len-2] = '\0';
	    fprintf(foutput, "%s(", nam);
	}
	else if (isfunc(fn))
	{
	    fprintf(foutput, "%s(", funcof(fn)->f_name->s_chars);
	}
	for (j = 0; j < callrecof(callstack->a_base[i])->nargs; ++j)
	{
	    object_t	*o = callrecof(callstack->a_base[i])->ap[j];

	    if (j > 0)
		fprintf(foutput, ", ");
	    if (isnull(o))
		fprintf(foutput, "NULL");
	    else if (isint(o))
		fprintf(foutput, "%ld", intof(o)->i_value);
	    else if (isfloat(o))
		fprintf(foutput, floatfmt, floatof(o)->f_value);
	    else if (isstring(o) || isregexp(o))
		pstrorreg(o, 0);
	    else
		fprintf(foutput, "[%s]", o->o_type->t_name);
	}
	fprintf(foutput, ")\n");
    }
}

static void
getscope(char *args)
{
    if (curscp == NULL)
	curscp = structof(v_top[-1]);
    reply(SCOPE, "%p", (void *)curscp);
}

static void
setscope(char *args)
{
    void	*p;
    
    if (sscanf(args, "%p", &p) != 1)
	reply(USER_ERROR, "can't recognise pointer '%s'", args);
    else if (p == NULL)
    {
	curscp = structof(v_top[-1]);
	reply(OKAY, "");
    }
    else if ((int)p & 3)
	reply(USER_ERROR, "invalid pointer");
    else if (!isstruct(objof(p)))
	reply(USER_ERROR, "not a struct");
    else
    {
	curscp = p;
	reply(OKAY, "");
    }
}

static void
scopeup(char *args)
{
    if (curscp == NULL)
	curscp = structof(v_top[-1]);
    if (curscp->s_super == NULL)
	reply(NOT_OKAY, "");
    else
    {
	curscp = curscp->s_super;
	reply(OKAY, "");
    }
}

static void
fetchcmd(char *arg)
{
    object_t	*key, *val;

    if (curscp == NULL)
	curscp = structof(v_top[-1]);
    if ((key = objof(new_cname(arg))) == NULL)
	reply(FETCHED_VALUE, "NULL");
    else
    {
	if ((val = fetch(curscp, key)) == objof(&o_null))
	    reply(FETCHED_VALUE, "NULL");
	else
	{
	    fprintf(foutput, "%c ", FETCHED_VALUE);
	    pobj(val, 0, 0);
	    fprintf(foutput, "\n");
	}
	reply(FETCHED_VALUE, "");
    }
    decref(key);
}

static void
assigncmd(char *arg)
{
    char	*p;
    object_t	*key;
    object_t	*val;

    if (curscp == NULL)
	curscp = structof(v_top[-1]);
    if ((p = strchr(arg, ' ')) == NULL)
	return;
    *p++ = '\0';
    if ((key = objof(new_cname(arg))) == NULL)
	return;
    if ((val = iexpr("%s", p)) == NULL)
	return;
    assign(curscp, key, val);
}

static void
scopekeys(char *arg)
{
    slot_t	*sl;

    (void)arg;
    if (curscp == NULL)
	curscp = structof(v_top[-1]);
    reply(SCOPE_KEYS, "%d", curscp->s_nels);
    for (sl = curscp->s_slots; sl - curscp->s_slots < curscp->s_nslots; ++sl)
    {
	if (sl->sl_key != NULL && sl->sl_value != NULL)
	{
	    pobj(sl->sl_key, 0, 0);
	    fprintf(foutput, "\n");
	}
    }
}

/*
 * Map command to cmd_t.
 */
static cmd_t *
search(char *p)
{
    cmd_t	*c;

    for (c = cmds; c - cmds < ncmds; ++c)
	if (*p == c->cmdname)
	    return c;
    return NULL;
}

/*
 * Debugger command loop - prompts user, reads input, calls command functions
 * etc...
 */
static void
stopped(void)
{
    char	cmdstr[256];
    char	*p;
    char	*q;
    cmd_t	*cmd;

    if (srcmrk == NULL || srcmrk->s_filename == startupfile)
	return;
    if (state != FINISHED)
	fprintf(foutput, "%c %s %d\n", STOPPED, srcmrk->s_filename->s_chars, srcmrk->s_lineno);
    for (;;)
    {
	if (fgets(cmdstr, sizeof cmdstr, finput) == NULL)
	    quit(NULL);
	if ((p = strchr(cmdstr, '\n')) == NULL)
	{
	    int	c;

	    fprintf(foutput, "Warning: line too long, ignoring.\n");
	    while ((c = fgetc(finput)) != EOF && c != '\n')
		;
	    continue;
	}
	*p = '\0';
	p = cmdstr;
	q = p + 1;
	*q++ = '\0';
	if ((cmd = search(p)) == NULL)
	    fprintf(foutput, "E unrecognised debugger command\n");
	else
	{
	    (*cmd->func)(q);
	    if (cmd->resume && state != BROKEN && state != FINISHED)
		break;
	}
    }
}

/**
 * Callback functions
 */

/*
 * ici_debug_error
 *
 * This is called when an error is raised in the program. We're passed
 * the error string and the source line marker where the error was
 * raised.
 */
void
ici_debug_error(char *err, src_t *src)
{
    init();
    if (dontstop)
	return;
    fprintf(foutput, "E %s\n", err);
    enterstate(BROKEN);
    stopped(); /* Never returns when state == BROKEN */
}

/*
 * ici_debug_fncall
 *
 * This is called when a function is about to be called. We're passed
 * the function object, which may be an ici func_t or a C cfunc_t, and
 * the array of objects being passed as parameters.
 */
void
ici_debug_fncall(object_t *o, object_t **ap, int nargs)
{
    callrec_t	*cr;

    init();
    if (dontstop)
	return;
    if (interrupt)
    {
	interrupt = 0;
	enterstate(SINGLE_STEP);
    }
#ifndef NDEBUG
    else if (state == FINISHED || state == BROKEN)
	fatal("internal error, bad debugger state");
#endif
    else if (state == RUN_TILL_BREAK && stopin(o))
	enterstate(SINGLE_STEP);
    if ((cr = new_callrec(o, ap, nargs)) == NULL || pushcheck(callstack, 1))
	fatal(error);
    *callstack->a_top++ = objof(cr);
    decref(cr);
}

/*
 * ici_debug_fnresult
 */
void
ici_debug_fnresult(object_t *o)
{
    callrec_t	*cr;

    init();
    if (dontstop)
	return;
    if (callstack->a_top > callstack->a_base)
	cr = callrecof(*--callstack->a_top);
    if (interrupt)
    {
	interrupt = 0;
	enterstate(SINGLE_STEP);
    }
#ifndef NDEBUG
    else if (state == FINISHED || state == BROKEN)
	fatal("internal error, bad debugger state");
#endif
    else if (state == RUN_TILL_RETURN && tos > callstack->a_top)
    {
	fprintf(foutput, "%c ", RETURNED_VALUE);
	pobj(o, 0, 0);
	fprintf(foutput, "\n");
	enterstate(SINGLE_STEP);
    }
}

/*
 * ici_debug_src
 *
 * This is called when a "source line marker" is encounted in the ici
 * operation stream. These are inserted into the code to tell us where
 * it comes from. We get the filename and the line number from this.
 */
void
ici_debug_src(src_t *src)
{
    init();
    if (dontstop)
	return;
    if (srcmrk != NULL)
	decref(srcmrk);
    srcmrk = src;
    incref(srcmrk);
    if (interrupt)
    {
	interrupt = 0;
	stopped();
    }
#ifndef NDEBUG
    else if (state == FINISHED || state == BROKEN)
	fatal("internal error, bad debugger state");
#endif
    else if
    (
	state == SINGLE_STEP
	||
	(state == SINGLE_STEP_OVER && tos >= callstack->a_top)
	||
	(state == RUN_TILL_BREAK && stopat(src))
    )
    {
	stopped();
    }
}

/*
 * ici_debug_watch
 */
void
ici_debug_watch(object_t *o, object_t *k, object_t *v)
{
    set_t	*set;

    init();
    if (dontstop)
	return;
    if (interrupt)
    {
	interrupt = 0;
	stopped();
    }
    else if (objof(set = setof(fetch(watched, o))) != objof(&o_null))
    {
	if (fetch(set, k) != objof(&o_null))
	{
	    reply(WATCHPOINT_HIT, NULL);
	    pobj(k, 0, 0);
	    fprintf(foutput, "\n");
	    pobj(fetch(o, k), 4, 4);
	    pobj(v, 4, 4);
	    fprintf(foutput, "\n");
	    stopped();
	}
    }
}

/*
 * Initialise things that need initialising. This is called via the
 * init() macro which hides the use of a flag to avoid calling this
 * more than once.
 */
static int
init_(void)
{
    if ((string__ = new_cname("_")) == NULL)
	goto fail;
    if ((string_id = new_cname("id")) == NULL)
	goto fail;
    if ((string_where = new_cname("where")) == NULL)
	goto fail;
    if ((string_Type = new_cname("type")) == NULL)
	goto fail;
    if ((string_func = new_cname("func")) == NULL)
	goto fail;
    if ((string_line = new_cname("line")) == NULL)
	goto fail;
    if ((breakfuncs = new_struct()) == NULL)
	goto fail;
    if ((breakpoints = new_struct()) == NULL)
	goto fail;
    if ((startupfile = new_cname("ici_init.ici")) == NULL)
	goto fail;
    if ((callstack = new_array()) == NULL)
	goto fail;
    if ((sources = new_struct()) == NULL)
	goto fail;
    if ((watched = new_struct()) == NULL)
	goto fail;
    return needinit = 0;

fail:
    return 1;
}
