/*
 * icid - a command line debugger for ici.
 * Andy Newman <atrn@zeta.org.au> or <andy@research.canon.com.au>
 *
 * This is a simple command line debugger front-end for ici. This
 * debugger provides the minimum facilities for actually debugging
 * a program. These are,
 *
 *	single-step, possibly skipping functions
 *	execute till end of function
 *	continue execution
 *	set a breakpoint (on a function or specific line)
 *	print an expression
 *	display stack traceback
 *
 * The debugger provides a display of the source code when the
 * program stops for some reason. Currently this is output with
 * printf however a version using a GUI is under development, also
 * a curses interface would not with too difficult.
 */

#define	VERSION\
 "icid -- command line ici debugger, version 0.2, Feb 27, 1999."

#define	COPYRIGHT\
 "Copyright (C) A.Newman <atrn@zeta.org.au>, 1999.\n"\
 "Free to use for any purpose as long as this notice is retained."

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <signal.h>
#include <ctype.h>

#include <object.h>
#include <int.h>
#include <float.h>
#include <str.h>
#include <re.h>
#include <array.h>
#include <set.h>
#include <struct.h>
#include <ptr.h>
#include <src.h>
#include <buf.h>
#include <func.h>
#include <exec.h>
#include <op.h>
#include <file.h>
#include <null.h>

/*
 * The different states we can be in. This is used to decide when we
 * should stop execution and prompt the user for what to do next. The
 * ici interpreter calls the debug callback functions at specific points
 * and these functions need to decide what to do according to the state.
 * (The reason for actually naming the enum and using this is so debuggers
 *  will be nice to us and print the enumerators, sad but true).
 */
enum state
{
    FINISHED,		/* Program has finished execution.		*/
    SINGLE_STEP,	/* Single-step, stop at each statement.		*/
    SINGLE_STEP_OVER,	/* Single-step, don't step through functions.	*/
    RUN_TILL_BREAK,	/* Run until a break- or watch-point is hit.	*/
    RUN_TILL_RETURN,	/* Run until the current function returns.	*/
    BROKEN		/* The program raised an unhandled error.	*/
};

/*
 * Global data...
 *
 *	dontstop
 *		This is set to true when ici is executing code on behalf
 *		of the debugger itself. The callbacks test this flag and
 *		immediately return if it is true to avoid infinite loops
 *		trying to debug the debugger.
 *
 *	state
 *		The current state of execution. This is used to control
 *		when the debugger stops and prompts the user. The callbacks
 *		make various decisions based on the current state.
 *
 *	watched
 *		Structure used to store watchpoints. The top-level struct
 *		maps scopes to sets of variables being watched within
 *		that scope. To enter a watchpoint we create a set for
 *		the current scope and add the key to watch to the set. To
 *		test a watchpoint we lookup the current scope's set and
 *		lookup the key in that set.
 *
 *	tos
 *		Value of the top of the callstack at particular times. This
 *		is used in the ``next'' and ``fin'' commands to determine
 *		just when we should stop - for next it's when we return to
 *		the same function call depth and for fin it's when we reduce
 *		the call nesting depth.
 *
 *	srcmrk
 *		The last source marker we've seen. This is set by the
 *		source marker callback function and used output and
 *		obtain the source code text and used when deciding when
 *		a breakpoint has been hit.
 *
 *	breakpoints
 *		A struct that maps a line number (int_t) to a struct
 *		that maps filenames (string_t) to structs describing
 *		the breakpoint. When testing for a breakpoint hit we
 *		look up the line number in the struct. If found we
 *		see if that line's struct has an entry for this file.
 *		It might be better to form an atomic array storing the
 *		filename and line number and use a single fetch. It's
 *		a pity that src_t's compare as unique always - we can't
 *		use them as keys. 'Tis a pity.
 *
 *	breakfuncs
 *		A struct mapping functions to structures describing
 *		the breakpoint on that function. This is used when
 *		a breakpoint is set on a function. On every call
 *		we lookup the function being called in this struct
 *		to decide if we need to break.
 *
 *	callstack
 *		Stack of records describing each function call
 *		in the current flow of control. Each record has
 *		a reference to the function object called and the
 *		parameter array passed to the function. See the
 *		callrec (ici) type defined below.
 *
 *	interrupt
 *		Set to non-zero when the interpreter should stop
 *		for a reason other than normal flow control. Incremented
 *		in the SIGINT signal handler. If a second interrupts occurs
 *		without the first being handled the debugger exits.
 *
 *	startupfile
 *		Ici string object version of the name of the interpreter
 *		startup file. We ignore this file.
 *
 *	needinit
 *		Non-zero if we need to call the initialisation code
 *		to create objects etc...
 *
 *	string_xxxx
 *		Cached reference to ici string object for xxx.
 *
 *	floatfmt
 *		The printf format used to output floating point variables.
 *		This can be set via the ``set'' command.
 *
 *	maxstrlen
 *		The maximum number of characters to output from an ici
 *		string object when printing it. If the string is longer
 *		than this an ellipsis is printed following maxstrlen
 *		characters.
 *
 *	maxarraylen
 *		Similar to maxstrlen. This controls how many elements
 *		of an array are printed. If the array has more elements
 *		that this an ellipsis is printed after the last object
 *		printed.
 *
 *	contextlines
 *		Number of lines either side of the current line listed
 *		when the debugger stops the program.
 *
 *	saystopped
 *		Flag that tells the stopped() function when to output
 *		the Stopped at line... message. This is only output
 *		on debug state transitions, e.g., when a breakpoint
 *		is hit. We don't want to output the message everytime
 *		we do something like another single-step command, we
 *		already know we're stopped.
 *
 *	breaknum
 *		Last breakpoint number.
 *
 */
FILE		*finput;
FILE		*foutput;
int		dontstop;
enum state	state = SINGLE_STEP;
struct_t	*watched;
object_t	**tos;
src_t		*srcmrk;
struct_t	*breakpoints;
struct_t	*breakfuncs;
array_t		*callstack;
volatile int	interrupt;
string_t	*startupfile;
int		needinit = 1;
struct_t	*sources;
string_t	*string__;
string_t	*string_id;
string_t	*string_where;
string_t	*string_Type;
string_t	*string_func;
string_t	*string_line;
char		*floatfmt = "%*s%.8g";
int		maxstrlen = 32;
int		maxarraylen = 3;
int		contextlines = 3;
int		saystopped = 1;
int		breaknum = 0;

#define	init()	 if (needinit && initf()) fatal(error)
#define	fatal(s) fatalf(s, __FILE__, __LINE__)
#define	min(a,b) ((a) < (b) ? (a) : (b))
#define	max(a,b) ((a) > (b) ? (a) : (b))

/*
 * Debugger commands are described by a cmd_t structure. It has the
 * command name and abbreviated form if it has one, a pointer to the
 * function that implements the command and two pieces of help text,
 * a short command description and a long-winded version.
 *
 * Command functions are passed a pointer to the text following the
 * command on the command line. They don't return any value.
 */
typedef struct
{
    char	*name;
    char	*abbrev;
    int		resume;
    void	(*func)(char *);
    char	*desc;
    char	*help;
}
cmd_t;

static void	autos(char *);
static void	breakp(char *);
static void	cont(char *);
static void	externs(char *);
static void	fin(char *);
static void	help(char *);
static void	info(char *);
static void	list(char *);
static void	next(char *);
static void	print(char *);
static void	quit(char *);
static void	set(char *);
static void	statics(char *);
static void	step(char *);
static void	version(char *);
static void	watch(char *);
static void	where(char *);

/*
 * All the commands are in this table.
 */
static cmd_t	cmds[] =
{
    {
	"autos",
	NULL,
	0,
	autos,
	"Display contents of the auto scope level",

	"The ``autos'' command displays the contents of the auto scope\n"
	"level. This is equivilent to the command ``print scope()''.\n"
	"To display other parts of the scope you may use the ``statics''\n"
	"and ``externs'' commands or specific print commands as shown."
    },
    {
	"break",
	"b",
	0,
	breakp,
	"Set a breakpoint at a function or line",

	"The ``breakpoint'' command is used to set breakpoints at either\n"
	"particular lines or upon calls to specific functions. To set a\n"
	"breakpoint at a particular line specify the breakpoint as\n"
	"<file:line> where ``file'' is the name of the file and ``line'' is\n"
	"the line you want to stop at."
    },
    {
	"cont",
	"c",
	1,
	cont,
	"Continue program execution",

	"The ``continue'' command is used to resume program execution\n"
	"when it stops for some reason. When continued, the program will\n"
	"execute until the next breakpoint or watchpoint is hit or until\n"
	"the program finishes."
    },
    {
	"externs",
	"e",
	0,
	externs,
	"Display the extern scope level",

	"The ``externs'' command lists the content of the extern scope\n"
	"level. The equivilent command is ``print super(super(scope()))''\n"
    },
    {
	"fin",
	"f",
	1,
	fin,
	"Run until the current function returns",

	"The ``finish'' command is used to continue execution until the\n"
	"the current function returns. Single-stepping resumes at the\n"
	"place where the function was called."
    },
    {
	"help",
	"h",
	0,
	help,
	"List commands understood by the debugger",

	"The ``help'' command may be used to either list all commands\n"
	"recognised by the debugger or to display information on specific\n"
	"commands."
    },
    {
	"info",
	"i",
	0,
	info,
	"List information on various things",

	"To display currently set breakpoints use,\n"
	"\n"
	"    info breakpoints\n"
	"\n"
	"To display currently set watchpoints use,\n"
	"\n"
	"    info watchpoints\n"
	"\n"
    },
    {
	"list",
	"l",
	0,
	list,
	"List source lines",

	"The ``list'' command lets you display source."
    },
    {
	"next",
	"n",
	1,
	next,
	"Run one statement stepping over functions",

	"To single-step but step-over function calls use the ``next''\n"
	"command."
    },
    {
	"print",
	"p",
	0,
	print,
	"Print the value of an expression",

	"The ``print'' command displays the result of evaluating an\n"
	"expression within the current scope. The most common use of\n"
	"the command is to display variable values but it may be used\n"
	"to execute any ici expression."
    },
    {
	"quit",
	"q",
	0,
	quit,
	"Exit the debugger",

	"Exit the debugger. If the program has not finished you are\n"
	"asked to confirm the command."
    },
    {
	"set",
	NULL,
	0,
	set,
	"Set option values",

	"Various debugger options may be set using the ``set'' command.\n"
    },
    {
	"statics",
	NULL,
	0,
	statics,
	"Display the static scope level",

	"The ``statics'' command displays the contents of the static scope\n"
	"level. It is equivilent to the command ``print super(scope())''."
    },
    {
	"step",
	"s",
	1,
	step,
	"Run one statement",

	"The ``step'' command runs the next statement and returns you to\n"
	"the debugger prompt."
    },
    {
	"version",
	"v",
	0,
	version,
	"Display debugger version information",

	"Display version and copyright information."
    },
    {
	"watch",
	NULL,
	0,
	watch,
	"Set a watch point on a variable",

	"The ``watch'' command sets a watch point on a particular variable.\n"
	"Watchpoints cause the program to stop and return to the debugger\n"
	"when that variable is modified."
    },
    {
	"where",
	"w",
	0,
	where,
	"Show function call tree",

	"The ``where'' command shows the current stack backtrace.\n"
	"Where possible function names are given and argument lists.\n"
	"There is currently a deficiency in the debugger which makes\n"
	"it show the current value of variables used as parameters to a\n"
	"function rather than the values used when the function was\n"
	"called. This will be corrected at a later date."
    }
};

#define ncmds (sizeof cmds/sizeof cmds[0])

/* More forward decls. */
static int	initf(void);
static void	stopped(void);
static void	sigint(int);
static void	fatalf(char *, char *, int);
static cmd_t	*search(char *);
static void	pobj(object_t *, int, int);
static void	enterstate(enum state);

void		ici_debug_error(char *, src_t *);
void		ici_debug_fncall(object_t *, object_t **, int);
void		ici_debug_fnresult(object_t *);
void		ici_debug_src(src_t *);
void		ici_debug_watch(object_t *, object_t *, object_t *);

debug_t	icid_debug_funcs =
{
    ici_debug_error,
    ici_debug_fncall,
    ici_debug_fnresult,
    ici_debug_src,
    ici_debug_watch
};

/*
 * Entry point... We just call the debug ici main function and tell
 * to enable debugging (the bare ``1''). This will result in our
 * callbacks being invoked when the code starts executing. If ici_maind
 * returns we go into finished mode.
 */
int
main(int argc, char **argv)
{
    int		i;

    ici_debug = &icid_debug_funcs;

    if (getenv("ICID_INFD") != NULL)
    {
	finput = fdopen(atoi(getenv("ICID_INFD")), "r");
	foutput = fdopen(atoi(getenv("ICID_OUTFD")), "w");
	setlinebuf(foutput);
	setlinebuf(finput);
	fprintf(stderr, "input %d output %d\n", atoi(getenv("ICID_INFD")), atoi(getenv("ICID_OUTFD")));
    }
    else
    {
	finput = stdin;
	foutput = stdout;
    }
    fprintf(foutput, "\nStarting program:");
    for (i = 1; i < argc; ++i)
	fprintf(foutput, " %s", argv[i]);
    fprintf(foutput, "\n");

    signal(SIGINT, sigint);
    i = ici_maind(argc, argv, 1);
    dontstop = 1;
    if (i == 0)
	fprintf(foutput, "\nProgram exited normally.\n\n");
    else
	fprintf(foutput, "\nProgram exited with code %0o\n\n", i);
    enterstate(FINISHED);
    stopped();
    return i;
}

#ifdef _WIN32

/*
 * The ANSI people provided a function "atexit" for registering other
 * functions to be called by exit(). Unfortunately, they forgot to
 * allow these functions to be passed the exit status, so we can't
 * easily implement what we want.  :-(
 *
 * ICI has "wraps" that are called at exit time and they may be used
 * to implement exit trapping. It may be better if the debugger simply
 * replaces ici's exit() with a special function that traps back to
 * the debugger.
 */

#else

/*
 * Version of exit() using _exit(). Flush all files before exiting.
 */
static void
my_exit(int rc)
{
    fflush(NULL);
    _exit(rc);
}

#ifdef __GNUC__
static void	my_exit(int) __attribute__ ((noreturn));
#endif

/*
 * This is gnarly. Override exit so we can trap ici calls to it.
 * The code is a copy of that in main() following the call to
 * ici_maind().
 */
void
exit(int rc)
{
    if (!dontstop)
    {
	if (rc == 0)
	    fprintf(foutput, "\nProgram exited normally.\n\n");
	else
	    fprintf(foutput, "\nProgram exited with code %0o\n\n", rc);
	enterstate(FINISHED);
	stopped();
    }
    my_exit(rc);
}

/*
 * Replace exit() with my_exit() in the remainder of this file so
 * we don't call the above function when we try to exit the
 * debugger itself.
 */
#define	exit(s)	my_exit(s)

#endif

/*
 * Trap SIGINT (and possibly others) and set the interrupt flag so
 * the callbacks know to stop the program. If more than one SIGINT
 * in a row exit immediately.
 */
static void
sigint(int sig)
{
    if (++interrupt > 1)
    {
	fprintf(foutput, "\n** Forced exit **\n");
	exit(1);
    }
}

/*
 * We define a new type, callrec, to record details of function calls.
 * We maintain an array of callrec objects to record the function call
 * "stack frames". This allows us to implement the "where" command.
 *
 * Each callrec contains the parameters to ici_debug_fncall (see below)
 * at the time of the function call. The argument pointer array, ap, is
 * a copy of that passed to ici_debug_fncall as the parameter points
 * directly into the ici stack which may change and has a "backwards"
 * layout in memory.
 *
 * XXX * The values the objects represent may change and the debugger
 * should really do a deep copy of the arguments so it can report the
 * initial values correctly. However I'll leave that out for now.
 */
typedef struct callrec
{
    object_t	o_head;
    object_t	*fn;
    object_t	**ap;
    int		nargs;
}
callrec_t;

#define	callrecof(o)	((callrec_t *)(o))

static unsigned long
mark_callrec(object_t *o)
{
    unsigned long	sz;
    int			i;

    o->o_flags |= O_MARK;
    for (sz = 0, i = 0; i < callrecof(o)->nargs; ++i)
	sz += mark(callrecof(o)->ap[i]);
    return sizeof (callrec_t) + mark(callrecof(o)->fn) + sz;
}

static type_t	callrec_type =
{
    mark_callrec,
    free_simple,
    hash_unique,
    cmp_unique,
    copy_simple,
    assign_simple,
    fetch_simple,
    "callrec"
};

static callrec_t *
new_callrec(object_t *fn, object_t **ap, int nargs)
{
    size_t	sz;
    callrec_t	*cr;

    sz = nargs * sizeof (object_t *);
    if ((cr = ici_alloc(sizeof (callrec_t) + sz)) != NULL)
    {
	int	i;

	objof(cr)->o_tcode = TC_OTHER;
	objof(cr)->o_flags = 0;
	objof(cr)->o_nrefs = 1;
	objof(cr)->o_type  = &callrec_type;
	rego(cr);
	cr->fn = fn;
	cr->ap = (object_t **)(cr + 1);
	cr->nargs = nargs;
	for (i = 0; i < nargs; ++i)
	{
	    cr->ap[i] = ap[-i]; /* Note: ap is backwards, see ici/func.h */
	    incref(cr->ap[i]);
	}
    }
    return cr;
}

/****
 * Utility functions...
 */

/*
 * Report a fatal error and exit. Used in conjuction with the fatal()
 * macro which fills in the filename and line number.
 */
static void
fatalf(char *msg, char *file, int line)
{
    fprintf(stderr, "FATAL ERROR: %s, line %d\n%s\n", file, line, msg);
    exit(1);
}

/*
 * Is this line (fgets buffer so it has the \n) a yes?
 */
static int
yes(char *s)
{
    return
	!strcmp(buf, "y\n")   || !strcmp(buf, "Y\n")
	||
	!strcmp(buf, "yes\n") || !strcmp(buf, "YES\n")
	||
	!strcmp(buf, "Yes\n");
}

/*
 * Prompt the user to see if they really want to continue when
 * the program has raised an error. Returns non-zero if they
 * want to continue, zero otherwise.
 */
static int
broken(void)
{
    fprintf
    (
	foutput,
	"\n"
	"The program has raised an error. Resuming execution\n"
	"may produce unpredictable results.\n"
	"\n"
	"Do you want to continue anyway ? "
    );
    fflush(foutput);
    if (fgets(buf, ici_bufz, finput) == NULL)
	*buf = '\0';
    if (yes(buf))
    {
	fprintf(foutput, "\nContinuing...\n\n");
	return 1;
    }
    fprintf(foutput, "\nNot continuing.\n");
    return 0;
}

/* 
 * Tell the user they can't do something because the program has finished.
 */
static void
finished(void)
{
    fprintf(foutput, "The program is no longer running.\n");
}

/*
 * This is set in the object flags if this object has been output.
 * We don't want to get into infinite loops when we output things.
 * That would be bad.
 */
#define	O_PRINT	0x20

/*
 * Output a string or regular expression object.
 */
static void
pstrorreg(object_t *o, int ind)
{
    int	i, n, c;

    n = min(maxstrlen, stringof(o)->s_nchars);
    fprintf(foutput, "%*s%c", ind, "", isstring(o) ? '"' : '#');
    for (i = 0; i < n; ++i)
    {
	switch (c = stringof(o)->s_chars[i])
	{
	case '\n':
	    fprintf(foutput, "\\n");
	    break;
	case '\r':
	    fprintf(foutput, "\\r");
	    break;
	case '\b':
	    fprintf(foutput, "\\b");
	    break;
	case '\v':
	    fprintf(foutput, "\\v");
	    break;
	case '\f':
	    fprintf(foutput, "\\f");
	    break;
	default:
	    if (c >= ' ' && c <= '~')
		putchar(c);
	    else
		fprintf(foutput, "\\x%02X", c);
	}
    }
    putchar(isstring(o) ? '"' : '#');
    if (stringof(o)->s_nchars > maxstrlen)
	fprintf(foutput, "...");
}

/*
 * Output a pointer
 */
static void
pptr(object_t *o, int ind1, int ind2)
{
    fprintf(foutput, "%*s&", ind1, "");
    pobj(ptrof(o)->p_aggr, 0, ind2);
    fprintf(foutput, "[");
    pobj(ptrof(o)->p_key, 0, ind2);
    fprintf(foutput, "]");
}

/*
 * Output an array
 */
static void
parray(object_t *o, int ind1, int ind2)
{
    object_t	**po;
    int		i;

    o->o_flags |= O_PRINT;
    fprintf(foutput, "%*s%s[array\n", ind1, "", o->o_flags & O_ATOM ? "@" : "");
    for
    (
	i = 0, po = arrayof(o)->a_base;
	i < maxarraylen && po < arrayof(o)->a_top;
	++po, ++i
    )
    {
	fprintf(foutput, "%*s[%d] = ", ind2 + 1, "", i);
	pobj(*po, 0, ind2 + 1);
	fprintf(foutput, ",\n");
    }
    if (po < arrayof(o)->a_top)
	fprintf(foutput, "%*s...\n", ind2 + 1, "");
    fprintf(foutput, "%*s]", ind2, "");
    o->o_flags &= ~O_PRINT;
}

/*
 * Output a structure
 */
static void
pstruct(object_t *o, int ind1, int ind2)
{
    slot_t	*sl;

    o->o_flags |= O_PRINT;
    fprintf(foutput, "%*s%s[struct\n", ind1, "", o->o_flags & O_ATOM ? "@" : "");
    for
    (
	sl = structof(o)->s_slots;
	(sl - structof(o)->s_slots) < structof(o)->s_nslots;
	++sl
    )
    {
	if (sl->sl_key != NULL)
	{
	    pobj(sl->sl_key, ind2 + 1, ind2 + 1);
	    fprintf(foutput, " = ");
	    pobj(sl->sl_value, 0, ind2 + 1);
	    fprintf(foutput, ",\n");
	}
    }
    fprintf(foutput, "%*s]", ind2, "");
    o->o_flags &= ~O_PRINT;
}

/*
 * Output a set.
 */
static void
pset(object_t *o, int ind1, int ind2)
{
    object_t	**po;

    o->o_flags |= O_PRINT;
    fprintf(foutput, "%*s%s[set\n", ind1, "", o->o_flags & O_ATOM ? "@" : "");
    for
    (
	po = setof(o)->s_slots;
	po - setof(o)->s_slots < setof(o)->s_nslots;
	++po
    )
    {
	if (*po != NULL)
	{
	    pobj(*po, ind2 + 1, ind2 + 1);
	    fprintf(foutput, ",\n");
	}
    }
    fprintf(foutput, "%*s]", ind2, "");
    o->o_flags &= ~O_PRINT;
}

#define	func_name(f) ((f)->f_name)

/*
 * Print an ici object to the standard output.
 */
static void
pobj(object_t *o, int ind1, int ind2)
{
    if (o->o_flags & O_PRINT)
	return;
    if (isnull(o))
	fprintf(foutput, "%*sNULL", ind1, "");
    else if (isint(o))
	fprintf(foutput, "%*s%ld", ind1, "", intof(o)->i_value);
    else if (isfloat(o))
	fprintf(foutput, floatfmt, ind1, "", floatof(o)->f_value);
    else if (isstring(o) || isregexp(o))
	pstrorreg(o, ind1);
    else if (isptr(o))
	pptr(o, ind1, ind2);
    else if (isarray(o))
	parray(o, ind1, ind2);
    else if (isset(o))
	pset(o, ind1, ind2);
    else if (isstruct(o))
	pstruct(o, ind1, ind2);
    else if (issrc(o))
	fprintf(foutput, "%s:%d", srcof(o)->s_filename->s_chars, srcof(o)->s_lineno);
    else if (isfunc(o))
	fprintf(foutput, "%*s%s()", ind1, "", func_name(funcof(o))->s_chars);
    else
	fprintf(foutput, "%*s[%s]", ind1, "", o->o_type->t_name);
}

/*
 * Execute ici code stored formed from a printf format string and args.
 */
static object_t *
iexpr(char *fmt, ...)
{
    va_list		vargs;
    char		fmtstr[1024];
    int			len;
    file_t		*file;
    int			rc;
    object_t		*o;

    if (chkbuf(1024 + strlen(fmt)))
	fatal(error);
    va_start(vargs, fmt);
    sprintf(fmtstr, "auto _=%s;\n", fmt);
    len = vsprintf(buf, fmtstr, vargs);
    va_end(vargs);
    if ((file = ici_sopen(buf, len)) == NULL)
	fatal(error);
    if ((file->f_name = get_cname(buf)) == NULL)
	fatal(error);
    dontstop = 1;
    rc = parse_module(file, structof(v_top[-1]));
    decref(file);
    dontstop = 0;
    if (rc < 0)
	return NULL;
    o = fetch(v_top[-1], string__);
    if (o == objof(&o_null) || o == NULL)
	return objof(&o_null);
    unassign_struct(structof(v_top[-1]), objof(string__));
    NEXT_VSVER;
    return o;
}

/*
 * Return the source text given a filename and line number.
 */
static char *
source(string_t *filename, int lineno)
{
    object_t		*o;

    if ((o = fetch(sources, filename)) == objof(&o_null))
    {
	char		*prog =
			"[func () {\n"
			"    auto f, a, s;\n"
			"    f = fopen(\"%s\");\n"
			"    a = array();\n"
			"    while (s = getline(f))\n"
			"	push(a, s);\n"
			"    return a;\n"
			"}]()";

	if ((o = iexpr(prog, filename->s_chars)) == NULL)
	    return "** Unable to access source code **";
	(void)assign(sources, filename, o);
    }
    if (lineno < 1 || lineno > arrayof(o)->a_top - arrayof(o)->a_base)
	return NULL;
    return stringof(arrayof(o)->a_base[lineno - 1])->s_chars;
}

/*
 * Output lines n through m of the file named in the src_t
 */
static void
listlines(src_t *src, int n, int m)
{
    char *s;

    for (; n <= m; ++n)
	if ((s = source(src->s_filename, n)) != NULL)
	    fprintf(foutput, "%s%s:%-d\t%s\n", n == src->s_lineno ? "+" : " ", src->s_filename->s_chars, n, s);
	else
	    fprintf(foutput, "\n");
}

/*
 * Set the debugger state with special action on transitions.
 */
static void
enterstate(enum state newstate)
{
    if (state != newstate)
	saystopped = 1;
    state = newstate;
}

/*
 * See if we should stop at the file/line described by the src_t.
 * Return non-zero if we should stop the program, zero otherwise.
 */
static int
stopat(src_t *src)
{
    object_t	*o;
    int_t	*i;
    int		rc;

    if ((i = new_int(src->s_lineno)) == NULL)
	fatal(error);
    o = fetch(breakpoints, i);
    decref(i);
    rc = !isnull(o);
    if (rc)
    {
	fprintf
	(
	    foutput,
	    "\n** Breakpoint hit at %s, line %d\n",
	    src->s_filename->s_chars,
	    src->s_lineno
	);
    }
    return rc;
}

/*
 * Determine if we should stop at a particular function.
 */
static int
stopin(object_t *fn)
{
    int		rc;

    if ((rc = !isnull(fetch(breakfuncs, fn))))
    {
	fprintf
	(
	    foutput,
	    "\n** Breakpoint hit on %s()\n",
	    func_name(funcof(fn))->s_chars
	);
    }
    return rc;
}

/****
 * Command functions...
 */

/*
 * Exit the debugger. Do whatever you need to do to clean up here.
 * At present there's nothing to do but that may change.
 */
static void
quit(char *arg)
{
    (void)arg;
    if (state != FINISHED && state != BROKEN)
    {
	fprintf(foutput, "\nThe program is running. Quit anyway ? ");
	fflush(stdout);
	if (fgets(buf, ici_bufz, finput) != NULL && !yes(buf))
	{
	    fprintf(foutput, "\n");
	    return;
	}
	fprintf(foutput, "\n");
    }
    exit(0);
}

/*
 * Print the result on an expression.
 */
static void
print(char *expr)
{
    static object_t	*prev = NULL;

    if (*expr != '\0')
	prev = iexpr(expr);
    else if (prev == NULL)
    {
	fprintf(foutput, "No previous print command.\n");
	return;
    }
    if (prev == NULL)
    {
	char	*p;

	/* Strip filename and line number off error text */
	if ((p = strchr(error, ':')) == NULL)
	    p = error;
	else
	    p += 2; /* There's a space following the colon */
	fprintf(foutput, "%s", p);
    }
    else
    {
	fprintf(foutput, "%s = ", expr);
	pobj(prev, 0, 0);
    }
    fprintf(foutput, "\n");
}

/*
 * Continue program execution.
 */
static void
cont(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
	finished();
    enterstate(RUN_TILL_BREAK);
}

/*
 * Single-step the program.
 */
static void
step(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
    {
	finished();
	return;
    }
    enterstate(SINGLE_STEP);
}

/*
 * List source lines
 */
static void
list(char *arg)
{
    if (*arg == '\0')
    {
	listlines
	(
	    srcmrk,
	    max(srcmrk->s_lineno - contextlines, 1),
	    srcmrk->s_lineno + contextlines
	);
    }
    else
    {
	int	first, last;

	switch (sscanf(arg, "%d,%d", &first, &last))
	{
	case 1:
	    last = first;
	case 2:
	    listlines(srcmrk, first, last);
	    break;
	default:
	    fprintf(foutput, "Invalid line range.\n");
	}
    }
}

/*
 * Single-step but don't trace through called functions.
 */
static void
next(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
    {
	finished();
	return;
    }
    tos = callstack->a_top;
    enterstate(SINGLE_STEP_OVER);
}

/*
 * Finish the current function.
 */
static void
fin(char *arg)
{
    (void)arg;
    if (state == BROKEN && broken())
	return;
    else if (state == FINISHED)
    {
	finished();
	return;
    }
    tos = callstack->a_top;
    enterstate(RUN_TILL_RETURN);
}

/*
 * Output version/copyright stuff.
 */
static void
version(char *arg)
{
    (void)arg;
    fprintf(foutput, "\n"	VERSION "\n" COPYRIGHT "\n\n");
}

/*
 * List all commands or help on a specific command.
 */
static void
help(char *what)
{
    cmd_t	*c;

    if (*what)
    {
	if ((c = search(what)) != NULL)
	    fprintf(foutput, "%s\t%s\n\n%s\n", what, c->desc, c->help);
	else
	    fprintf(foutput, "No help for ``%s''.\n", what);
    }
    else
    {
	fprintf(foutput, "\nRecognised commands are\n\n");
	for (c = cmds; c - cmds < ncmds; ++c)
	{
	    if (c->abbrev != NULL)
		fprintf(foutput, "%-10s (%s)", c->name, c->abbrev);
	    else
		fprintf(foutput, "%-14s", c->name);
	    fprintf(foutput, "\t%s\n", c->desc);
	}
	fprintf(foutput, "\nLetters in parentheses are short forms of commands.\n\n");
    }
}

static struct_t *
breakp_line(char *filename, int lineno)
{
    int_t	*line;
    struct_t	*s1;
    struct_t	*s2;
    string_t	*n;

    if ((line = new_int(lineno)) == NULL)
	fatal(error);
    if ((n = new_cname(filename)) == NULL)
	fatal(error);
    if ((s2 = structof(fetch(breakpoints, line))) == structof(&o_null))
    {
	if ((s2 = new_struct()) == NULL)
	    fatal(error);
	if (assign(breakpoints, line, s2))
	    fatal(error);
	decref(s2);
    }
    else if (fetch(s2, n) != objof(&o_null))
    {
	fprintf(foutput, "Breakpoint already set at %s:%d\n", filename, lineno);
	decref(line);
	decref(n);
	return NULL;
    }
    if ((s1 = new_struct()) == NULL)
	fatal(error);
    if (assign(s2, n, s1))
	fatal(error);
    decref(s1);
    decref(n);
    return s1;
}

/*
 * Set a breakpoint at a particular file/line or on a function.
 */
static void
breakp(char *where)
{
    char	*p;
    object_t	*o;
    int		lineno;
    struct_t	*s = 0;
    string_t	*n;
    int_t	*i;
    string_t	*type;

    /*
     * Do a reverse search to cope with `:'s in filenames.
     */
    if ((p = strrchr(where, ':')) == NULL)
    {
	if ((n = new_cname(where)) == NULL)
	    fatal(error);
	if ((o = fetch(v_top[-1], n)) == objof(&o_null))
	{
	    fprintf(foutput, "%s undefined.\n", where);
	    return;
	}
	if (!isfunc(o))
	{
	    fprintf(foutput, "%s is not a function.\n", where);
	    return;
	}
	if ((s = new_struct()) == NULL)
	    fatal(error);
	if (assign(breakfuncs, o, s))
	    fatal(error);
	type = string_func;
    }
    else
    {
	if (sscanf(p+1, "%u", &lineno) != 1 || lineno < 1)
	{
	    fprintf(foutput, "Invalid breakpoint specification.\n");
	    return;
	}
	if ((s = breakp_line(where, lineno)) == NULL)
	    return;
	type = string_line;
    }
    decref(s);
    if ((i = new_int(++breaknum)) == NULL || assign(s, string_id, i))
	fatal(error);
    decref(i);
    if ((n = new_cname(where)) == NULL || assign(s, string_where, n))
	fatal(error);
    decref(n);
    if (assign(s, string_Type, type))
	fatal(error);
    fprintf(foutput, "Breakpoint %d set on %s", breaknum, where);
    if (p != NULL)
	fprintf(foutput, ", line %d", lineno);
    fprintf(foutput, "\n");
}

/*
 * Set a watchpoint on a variable.
 */
static void
watch(char *arg)
{
    string_t	*s;
    set_t	*set;
    struct_t	*ss;

    if ((s = new_cname(arg)) == NULL)
	fatal(error);
    ss = structof(v_top[-1]);
    if (find_slot(&ss, objof(s))->sl_key != objof(s))
    {
	fprintf(foutput, "No such variable: %s.\n", arg);
	decref(s);
	return;
    }
    if (objof(set = setof(fetch(watched, v_top[-1]))) == objof(&o_null))
    {
	if ((set = new_set()) == NULL || assign(watched, v_top[-1], set))
	    fatal(error);
	decref(set);
    }
    if (assign(set, s, o_one))
	fatal(error);
    decref(s);
}

/*
 * Output the function call stack.
 */
static void
where(char *arg)
{
    int		i, j;

    (void)arg;
    if (state == FINISHED)
    {
	finished();
	return;
    }
    if (callstack == NULL || callstack->a_top == callstack->a_base)
    {
	fprintf(foutput, "Not within a function call.\n");
	return;
    }
    for (i = 0; i < callstack->a_top - callstack->a_base; ++i)
    {
	fprintf
	(
	    foutput,
	    "%*s%d\t%*s%s(",
	    i,
	    "",
	    i + 1,
	    i,
	    "",
	    func_name(funcof(callrecof(callstack->a_base[i])->fn))->s_chars
	);
	for (j = 0; j < callrecof(callstack->a_base[i])->nargs; ++j)
	{
	    object_t	*o = callrecof(callstack->a_base[i])->ap[j];

	    if (j > 0)
		fprintf(foutput, ", ");
	    if (isnull(o))
		fprintf(foutput, "NULL");
	    else if (isint(o))
		fprintf(foutput, "%ld", intof(o)->i_value);
	    else if (isfloat(o))
		fprintf(foutput, floatfmt, floatof(o)->f_value);
	    else if (isstring(o) || isregexp(o))
		pstrorreg(o, 0);
	    else
		fprintf(foutput, "[%s]", o->o_type->t_name);
	}
	fprintf(foutput, ")\n");
    }
}

/*
 * Print information about breakpoints.
 */
static void
info_break(void)
{
    static object_t	*fn;
    struct slot		*sl;
    struct slot		*sl2;
    array_t		*a;
    struct_t		*s;
    object_t		**po;
    object_t		*o;

    if (breakpoints->s_nels == 0 && breakfuncs->s_nels == 0)
    {
	fprintf(foutput, "No breakpoints defined.\n");
	return;
    }
    if ((a = new_array()) == NULL)
	fatal(error);
    if (breakpoints->s_nels > 0)
    {
	for
	(
	    sl = breakpoints->s_slots;
	    sl - breakpoints->s_slots < breakpoints->s_nslots;
	    ++sl
	)
	{
	    if (sl->sl_key == NULL)
		continue;
	    s = structof(sl->sl_value);
	    for (sl2 = s->s_slots; sl2 - s->s_slots < s->s_nslots; ++sl2)
	    {
		if (sl2->sl_key == NULL)
		    continue;
		if (pushcheck(a, 1))
		    fatal(error);
		*a->a_top++ = sl2->sl_value;
	    }
	}
    }
    if (breakfuncs->s_nels > 0)
    {
	for
	(
	    sl = breakfuncs->s_slots;
	    sl - breakfuncs->s_slots < breakfuncs->s_nslots;
	    ++sl
	)
	{
	    if (sl->sl_key != NULL)
	    {
		if (pushcheck(a, 1))
		    fatal(error);
		*a->a_top++ = sl->sl_value;
	    }
	}
    }
    if (fn == NULL)
    {
	fn = iexpr
	(
	    "[func (a, b)"
	    "{"
	    "    if (a.id < b.id)"
	    "        return -1;"
	    "    return a.id > b.id;"
	    "}]"
	);
	if (fn == NULL)
	    fatal(error);
	incref(fn);
    }
    dontstop = 1;
    ici_call("sort", "oo", a, fn);
    dontstop = 0;
    fprintf(foutput, "\nBreakpoints\n\n");
    for (po = a->a_base; po < a->a_top; ++po)
    {
	if ((o = fetch(*po, string_id)) == objof(&o_null))
	    fatal(error);
	fprintf(foutput, " %-3ld ", intof(o)->i_value);
	if ((o = fetch(*po, string_where)) == objof(&o_null))
	    fatal(error);
	fprintf(foutput, "%s", stringof(o)->s_chars);
	if (fetch(*po, string_Type) == objof(string_func))
	    fprintf(foutput, "()");
	fprintf(foutput, "\n");
    }
    fprintf(foutput, "\n");
}

static void
info_watch(void)
{
    if (watched->s_nels == 0)
	fprintf(foutput, "No watchpoints defined.\n");
    else
    {
    }
}

static struct info
{
    char	*name;
    void	(*func)(void);
    char	*desc;
}
infos[] =
{
    {"break",	info_break,	"List breakpoints"},
    {"watch",	info_watch,	"List watchpoints"}
};

#define	ninfos (sizeof infos/sizeof infos[0])

/*
 * Provide information on various things.
 */
static void
info(char *what)
{
    struct info	*info;

    if (*what == '\0')
    {
	fprintf(foutput, "\nThe following info commands exist,\n\n");
	for (info = infos; info - infos < ninfos; ++info)
	    fprintf(foutput, "  info %s -- %s\n", info->name, info->desc);
	fprintf(foutput, "\n");
    }
    else
    {
	for (info = infos; info - infos < ninfos; ++info)
	    if (!strcmp(what, info->name))
		break;
	if (info - infos == ninfos)
	    fprintf(foutput, "Don't know anything about \"%s\".\n", what);
	else
	    (*info->func)();
    }
}

/*
 * Set the various settable things.
 */
static void
set(char *args)
{
    char	*p, *q;
    int		i;

    static struct
    {
	char	*name;
	char	type;
	void	*addr;
    }
    vars[] =
    {
	{"arraylen",  'i', &maxarraylen},
	{"stringlen", 'i', &maxstrlen},
	{"context",   'i', &contextlines}
    };

#define nvars (sizeof vars/sizeof vars[0])

    for (p = args; isspace((int)*p); ++p)
	;
    for (q = p; *q && !isspace((int)*q); ++q)
	;
    if (*q)
	*q = '\0';
    if (q == p || !strcmp(p, "?"))
    {
	for (i = 0; i < nvars; ++i)
	{
	    fprintf(foutput, "%s\t", vars[i].name);
	    switch (vars[i].type)
	    {
	    case 'i':
		fprintf(foutput, "%d", *(int *)vars[i].addr);
		break;

	    case 's':
		fprintf(foutput, "%s", (char *)vars[i].addr);
		break;

	    default:
		fatal("internal error, bad type code in vars table");
	    }
	    fprintf(foutput, "\n");
	}
    }
    else
    {
	for (i = 0; i < nvars; ++i)
	{
	    if (!strcmp(p, vars[i].name))
	    {
		int	tmpi;

		++q;
		switch (vars[i].type)
		{
		case 'i':
		    if (sscanf(q, "%d", &tmpi) == 1)
			*(int *)vars[i].addr = tmpi;
		    else
			fprintf(foutput, "Bad integer value.\n");
		    break;
		}
	    }
	}
    }
}

/*
 * Print an individual scope level.
 */
static void
pscp(struct_t *scp)
{
    if (scp == NULL)
	pobj(objof(&o_null), 0, 0);
    else
	pobj(objof(scp), 0, 0);
    fprintf(foutput, "\n");
}

/*
 * Output the auto scope level.
 */
static void
autos(char *arg)
{
    (void)arg;
    pscp(structof(v_top[-1]));
}

/*
 * Output the static scope level.
 */
static void
statics(char *arg)
{
    (void)arg;
    pscp(structof(v_top[-1])->s_super);
}

/*
 * Output the extern scope level.
 */
static void
externs(char *arg)
{
    struct_t	*s;

    (void)arg;
    if ((s = structof(v_top[-1])->s_super) == NULL)
	pscp(NULL);
    else
	pscp(s->s_super);
}


/****
 * Debugger proper
 */

/*
 * Map command name or abbreviation to its cmd_t.
 */
static cmd_t *
search(char *p)
{
    cmd_t	*c;

    for (c = cmds; (c - cmds) < ncmds; ++c)
    {
	if (!strcmp(p, c->name))
	    return c;
	if (c->abbrev != NULL && !strcmp(p, c->abbrev))
	    return c;
    }
    return NULL;
}

/*
 * Debugger command loop - prompts user, reads input, calls command functions
 * etc...
 */
static void
stopped(void)
{
    char	cmdstr[256];
    static char	lastcmd[256] = "";
    char	*p;
    char	*q;
    cmd_t	*cmd;

    if (srcmrk == NULL || srcmrk->s_filename == startupfile)
	return;
    if (state != FINISHED)
    {
	fprintf(foutput, "\n");
	/* Only on state transitions - see enterstate() */
	if (saystopped)
	{
	    fprintf
	    (
		foutput,
		"Stopped at %s, line %d\n\n",
		srcmrk->s_filename->s_chars,
		srcmrk->s_lineno
	    );
	    saystopped = 0;
	}
	listlines
	(
	    srcmrk,
	    max(srcmrk->s_lineno - contextlines, 1),
	    srcmrk->s_lineno + contextlines
	);
	fprintf(foutput, "\n");
    }
    for (;;)
    {
	fprintf(foutput, "icid> ");
	fflush(stdout);
	if (fgets(cmdstr, sizeof cmdstr, finput) == NULL)
	    quit(0);
	if ((p = strchr(cmdstr, '\n')) == NULL)
	{
	    int	c;

	    fprintf(foutput, "Warning: line too long, ignoring it...\n");
	    while ((c = fgetc(finput)) != EOF && c != '\n')
		;
	    continue;
	}
	*p = '\0';
	if (cmdstr[0] != '\0')
	    strcpy(lastcmd, cmdstr);
	else if (lastcmd[0] != '\0')
	    strcpy(cmdstr, lastcmd);
	else /* Default command */
	    strcpy(cmdstr, "step");
	for (p = cmdstr; isspace((int)*p); ++p)
	    ;
	for (q = p; *q && !isspace((int)*q); ++q)
	    ;
	*q++ = '\0';
	if (!strcmp(p, "?"))
	    p = "help";
	if ((cmd = search(p)) == NULL)
#if 0
	    fprintf(foutput, "Unrecognised command: \"%s\"\n", p);
#else
	    fprintf(foutput, "What you talkin' 'bout Willis?\n");
#endif
	else
	{
	    (*cmd->func)(q);
	    if (cmd->resume && state != BROKEN && state != FINISHED)
		break;
	}
    }
}

/**
 * Callback functions
 */

/*
 * ici_debug_error
 *
 * This is called when an error is raised in the program. We're passed
 * the error string and the source line marker where the error was
 * raised.
 */
void
ici_debug_error(char *err, src_t *src)
{
    init();
    if (dontstop)
	return;
    fprintf(foutput, "\n** ERROR: %s\n\n", err);
    enterstate(BROKEN);
    stopped(); /* Never returns when state == BROKEN */
}

/*
 * ici_debug_fncall
 *
 * This is called when a function is about to be called. We're passed
 * the function object, which may be an ici func_t or a C cfunc_t, and
 * the array of objects being passed as parameters.
 */
void
ici_debug_fncall(object_t *o, object_t **ap, int nargs)
{
    callrec_t	*cr;

    init();
    if (dontstop)
	return;
    if (interrupt)
    {
	interrupt = 0;
	enterstate(SINGLE_STEP);
    }
#ifndef NDEBUG
    else if (state == FINISHED || state == BROKEN)
	fatal("internal error, bad debugger state");
#endif
    else if (state == RUN_TILL_BREAK && stopin(o))
	enterstate(SINGLE_STEP);
    if ((cr = new_callrec(o, ap, nargs)) == NULL || pushcheck(callstack, 1))
	fatal(error);
    *callstack->a_top++ = objof(cr);
}

/*
 * ici_debug_fnresult
 */
void
ici_debug_fnresult(object_t *o)
{
    callrec_t	*cr;

    init();
    if (dontstop)
	return;
    if (callstack->a_top > callstack->a_base)
    {
	cr = callrecof(*--callstack->a_top);
	decref(cr);
    }
    if (interrupt)
    {
	interrupt = 0;
	enterstate(SINGLE_STEP);
    }
#ifndef NDEBUG
    else if (state == FINISHED || state == BROKEN)
	fatal("internal error, bad debugger state");
#endif
    else if (state == RUN_TILL_RETURN && tos > callstack->a_top)
    {
	fprintf(foutput, "\nValue returned = ");
	pobj(o, 0, 0);
	fprintf(foutput, "\n");
	enterstate(SINGLE_STEP);
    }
}

/*
 * ici_debug_src
 *
 * This is called when a "source line marker" is encounted in the ici
 * operation stream. These are inserted into the code to tell us where
 * it comes from. We get the filename and the line number from this.
 */
void
ici_debug_src(src_t *src)
{
    init();
    if (dontstop || src == srcmrk)
	return;
    if (srcmrk != NULL)
	decref(srcmrk);
    srcmrk = src;
    incref(srcmrk);
    if (interrupt)
    {
	interrupt = 0;
	stopped();
    }
#ifndef NDEBUG
    else if (state == FINISHED || state == BROKEN)
	fatal("internal error, bad debugger state");
#endif
    else if
    (
	state == SINGLE_STEP
	||
	(state == SINGLE_STEP_OVER && tos >= callstack->a_top)
	||
	(state == RUN_TILL_BREAK && stopat(src))
    )
    {
	stopped();
    }
}

/*
 * ici_debug_watch
 */
void
ici_debug_watch(object_t *o, object_t *k, object_t *v)
{
    set_t	*set;

    init();
    if (dontstop)
	return;
    if (interrupt)
    {
	interrupt = 0;
	stopped();
    }
    else if (objof(set = setof(fetch(watched, o))) != objof(&o_null))
    {
	if (fetch(set, k) != objof(&o_null))
	{
	    fprintf(foutput, "\nWatchpoint hit on ");
	    pobj(k, 0, 0);
	    fprintf(foutput, "\n");
	    fprintf(foutput, "Current value\n");
	    pobj(fetch(o, k), 4, 4);
	    fprintf(foutput, "\nValue being assigned\n");
	    pobj(v, 4, 4);
	    fprintf(foutput, "\n");
	    stopped();
	}
    }
}

/*
 * Initialise things that need initialising. This is called via the
 * init() macro which hides the use of a flag to avoid calling this
 * more than once.
 */
static int
initf(void)
{
    if ((string__ = new_cname("_")) == NULL)
	goto fail;
    if ((string_id = new_cname("id")) == NULL)
	goto fail;
    if ((string_where = new_cname("where")) == NULL)
	goto fail;
    if ((string_Type = new_cname("type")) == NULL)
	goto fail;
    if ((string_func = new_cname("func")) == NULL)
	goto fail;
    if ((string_line = new_cname("line")) == NULL)
	goto fail;
    if ((breakfuncs = new_struct()) == NULL)
	goto fail;
    if ((breakpoints = new_struct()) == NULL)
	goto fail;
    if ((startupfile = new_cname("ici_init.ici")) == NULL)
	goto fail;
    if ((callstack = new_array()) == NULL)
	goto fail;
    if ((sources = new_struct()) == NULL)
	goto fail;
    if ((watched = new_struct()) == NULL)
	goto fail;
    return needinit = 0;

fail:
    return 1;
}
